# -*- coding: utf-8 -*-
# Copyright (C) Duncan Macleod (2013)
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""Definition of a BodePlot
"""

from math import pi
import numpy

from scipy import signal
from matplotlib.ticker import MultipleLocator

from .core import Plot
from ..spectrum import Spectrum

from .. import version
__version__ = version.version
__author__ = "Duncan Macleod <duncan.macleod@ligo.org>"
__all__ = ['BodePlot']


def to_db(a):
    """Convert the input array into decibels

    Parameters
    ----------
    a : `float`, `numpy.ndarray`
        value or array of values to convert to decibels

    Returns
    -------
    dB : ``10 * numpy.log10(a)``
    """
    return 10 * numpy.log10(a)


class BodePlot(Plot):
    """A `Plot` class for visualising transfer functions.

    Parameters
    ----------
    *filters : `~scipy.signal.lti`, `~gwpy.spectrum.Spectrum`, `tuple`
        any number of the following:

        - linear time-invariant filters, either
          `~scipy.signal.lti` or `tuple` of the following length and form:
             - 2: (numerator, denominator)
             - 3: (zeros, poles, gain)
             - 4: (A, B, C, D)

        - complex-valued `spectra <gwpy.spectrum.Spectrum>` representing
          a transfer function

    frequencies : `numpy.ndarray`, optional
        list of frequencies (in Hertz) at which to plot
    db : `bool`, optional, default: `True`
        if `True`, display magnitude in decibels, otherwise display
        amplitude.
    **kwargs
        other keyword arguments as applicable for `Plot` or
        :meth:`~SpectrumAxes.plot`

    Returns
    -------
    plot : `BodePlot`
        a new BodePlot with two `Axes` - :attr:`~BodePlot.maxes` and
        :attr:`~BodePlot.paxes` - representing the magnitude and
        phase of the input transfer function(s) respectively.
    """
    def __init__(self, *filters, **kwargs):
        """Initialise a new TimeSeriesPlot
        """
        dB = kwargs.pop('dB', True)
        frequencies = kwargs.pop('frequencies', None)

        # parse plotting arguments
        figargs = dict()
        for key in ['figsize', 'dpi', 'facecolor', 'edgecolor', 'linewidth',
                    'frameon', 'subplotpars', 'tight_layout']:
            if key in kwargs:
                figargs[key] = kwargs.pop(key)

        # generate figure
        super(BodePlot, self).__init__(**figargs)

        # delete the axes, and create two more
        self.add_subplot(2, 1, 1, projection='spectrum')
        self.add_subplot(2, 1, 2, projection='spectrum', sharex=self.maxes)

        # add filters
        for filter_ in filters:
            if isinstance(filter_, Spectrum):
                self.add_spectrum(filter_, dB=dB, **kwargs)
            else:
                self.add_filter(filter_, frequencies=frequencies,
                                dB=dB, **kwargs)

        # format plots
        if dB:
            self.maxes.set_ylabel('Magnitude [dB]')
        else:
            self.maxes.set_yscale('log')
            self.maxes.set_ylabel('Amplitude')
        self.paxes.set_xlabel('Frequency [Hz]')
        self.paxes.set_ylabel('Phase [deg]')
        self.maxes.set_xscale('log')
        self.paxes.set_xscale('log')
        self.paxes.yaxis.set_major_locator(MultipleLocator(base=90))
        self.paxes.set_ylim(-180, 180)

        # get xlim
        if (frequencies is None and len(filters) == 1 and
                isinstance(filters[0], Spectrum)):
            frequencies = filters[0].frequencies.value
        if frequencies is not None:
            frequencies = frequencies[frequencies > 0]
            self.maxes.set_xlim(frequencies.min(), frequencies.max())

    @property
    def maxes(self):
        """`SpectrumAxes` for the Bode magnitude
        """
        return self.axes[0]

    @property
    def paxes(self):
        """`SpectrumAxes` for the Bode phase
        """
        return self.axes[1]

    def add_filter(self, filter_, frequencies=None, dB=True, **kwargs):
        """Add a linear time-invariant filter to this BodePlot

        Parameters
        ----------
        filter_ : `~scipy.signal.lti`, `tuple`
            the filter to plot, either as a `~scipy.signal.lti`, or a
            `tuple` with the following number and meaning of elements

                 - 2: (numerator, denominator)
                 - 3: (zeros, poles, gain)
                 - 4: (A, B, C, D)

        frequencies : `numpy.ndarray`, optional
            list of frequencies (in Hertz) at which to plot
        db : `bool`, optional, default: `True`
            if `True`, display magnitude in decibels, otherwise display
            amplitude.
        **kwargs
            any other keyword arguments accepted by
            :meth:`~matplotlib.axes.Axes.plot`

        Returns
        -------
        mag, phase : `tuple` of `lines <matplotlib.lines.Line2D>`
            the lines drawn for the magnitude and phase of the filter.
        """
        if frequencies is None:
            w = None
        else:
            w = frequencies * 2. * pi
        if not isinstance(filter_, signal.lti):
            filter_ = signal.lti(*filter_)
        w, h = signal.freqs(filter_.num, filter_.den, w)
        w /= (2. * pi)
        mag = numpy.absolute(h)
        if dB:
            mag = 2 * to_db(mag)
        phase = numpy.degrees(numpy.unwrap(numpy.angle(h)))
        lm = self.maxes.plot(w, mag, **kwargs)
        lp = self.paxes.plot(w, phase, **kwargs)
        return lm, lp

    def add_spectrum(self, spectrum, dB=True, power=False, **kwargs):
        """Plot the magnitude and phase of a complex-valued Spectrum

        Parameters
        ----------
        spectrum : `~gwpy.spectrum.Spectrum`
            the (complex-valued) `Spectrum` to display
        db : `bool`, optional, default: `True`
            if `True`, display magnitude in decibels, otherwise display
            amplitude.
        power : `bool`, optional, default: `False`
            given `True` to incidate that ``spectrum`` holds power values,
            so ``dB = 10 * log(abs(spectrum))``, otherwise
            ``db = 20 * log(abs(spectrum))``. This argument is ignored if
            ``db=False``.
        **kwargs
            any other keyword arguments accepted by
            :meth:`~matplotlib.axes.Axes.plot`

        Returns
        -------
        mag, phase : `tuple` of `lines <matplotlib.lines.Line2D>`
            the lines drawn for the magnitude and phase of the filter.
        """
        # parse spectrum arguments
        kwargs.setdefault('label', spectrum.name)
        # get magnitude
        mag = numpy.absolute(spectrum.value)
        if dB:
            mag = to_db(mag)
            if not power:
                mag *= 2.
        # get phase
        phase = numpy.angle(spectrum.value, deg=True)
        # plot
        w = spectrum.frequencies.value
        lm = self.maxes.plot(w, mag, **kwargs)
        lp = self.paxes.plot(w, phase, **kwargs)
        return lm, lp



# -*- coding: utf-8 -*-
# Copyright (C) Duncan Macleod (2013)
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""Read and write HDF5 files in the LIGO Open Science Center format

For more details, see https://losc.ligo.org
"""

from glue.lal import CacheEntry

from astropy.io import registry
from astropy.units import (Unit, Quantity)

from .. import (StateVector, TimeSeries, TimeSeriesList)
from ...utils.deps import with_import
from ...io.cache import file_list
from ...io.hdf5 import open_hdf5


def read_losc_data(filename, channel, group=None, copy=False):
    """Read a `TimeSeries` from a LOSC-format HDF file.

    Parameters
    ----------
    filename : `str`
        path to LOSC-format HDF5 file to read.
    channel : `str`
        name of HDF5 dataset to read.
    group : `str`, optional
        name of containing HDF5 group for ``channel``. If not given,
        the first dataset named ``channel`` will be assumed as the right
        one.
    start : `Time`, `~gwpy.time.LIGOTimeGPS`, optional
        start GPS time of desired data
    end : `Time`, `~gwpy.time.LIGOTimeGPS`, optional
        end GPS time of desired data

    Returns
    -------
    data : :class`~gwpy.timeseries.TimeSeries`
        a new `TimeSeries` containing the data read from disk
    """
    h5file = open_hdf5(filename)
    if group:
        channel = '%s/%s' % (group, channel)
    dataset = _find_dataset(h5file, channel)
    # read data
    nddata = dataset.value
    # read metadata
    xunit = Unit(dataset.attrs['Xunits'])
    epoch = dataset.attrs['Xstart']
    dt = Quantity(dataset.attrs['Xspacing'], xunit)
    unit = Unit(dataset.attrs['Yunits'])
    # build and return
    return TimeSeries(nddata, epoch=epoch, sample_rate=(1/dt).to('Hertz'),
                      unit=unit, name=channel.rsplit('/', 1)[0], copy=copy)


def read_losc_data_cache(f, channel, start=None, end=None, resample=None,
                         group=None, target=TimeSeries):
    """Read a `TimeSeries` from a LOSC-format HDF file.

    Parameters
    ----------
    source : `str`, `list`, `~glue.lal.Cache`
        path to LOSC-format HDF5 file to read or cache of many files.
    channel : `str`
        name of HDF5 dataset to read.
    group : `str`, optional
        name of containing HDF5 group for ``channel``. If not given,
        the first dataset named ``channel`` will be assumed as the right
        one.
    start : `Time`, `~gwpy.time.LIGOTimeGPS`, optional
        start GPS time of desired data
    end : `Time`, `~gwpy.time.LIGOTimeGPS`, optional
        end GPS time of desired data

    Returns
    -------
    data : :class`~gwpy.timeseries.TimeSeries`
        a new `TimeSeries` containing the data read from disk
    """
    files = file_list(f)

    out = None
    for fp in files:
        if target is TimeSeries:
            new = read_losc_data(fp, channel, group=group, copy=False)
        elif target is StateVector:
            new = read_losc_state(fp, channel, group=group, copy=False)
        else:
            raise ValueError("Cannot read %s from LOSC data"
                             % (target.__name__))
        if out is None:
            out = new.copy()
        else:
            out.append(new)

    if resample:
        out = out.resample(resample)

    if start or end:
        out = out.crop(start=start, end=end)

    return out


def read_losc_state(filename, channel, group=None, start=None, end=None,
                    copy=False):
    """Read a `StateVector` from a LOSC-format HDF file.
    """
    h5file = open_hdf5(filename)
    if group:
        channel = '%s/%s' % (group, channel)
    # find data
    dataset = _find_dataset(h5file, '%s/DQmask' % channel)
    maskset = _find_dataset(h5file, '%s/DQDescriptions' % channel)
    # read data
    nddata = dataset.value
    bits = list(maskset.value)
    # read metadata
    try:
        epoch = dataset.attrs['Xstart']
    except KeyError:
        try:
            ce = CacheEntry.from_T050017(h5file.filename)
        except ValueError:
            epoch = None
        else:
            epoch = ce.segment[0]
    try:
        dt = dataset.attrs['Xspacing']
    except KeyError:
        dt = Quantity(1, 's')
    else:
        xunit = Unit(dataset.attrs['Xunit'])
        dt = Quantity(dt, xunit)
    return StateVector(nddata, bits=bits, epoch=epoch, name='Data quality',
                       sample_rate=(1/dt).to('Hertz'), copy=copy)


def read_losc_state_cache(*args, **kwargs):
    """Read a `StateVector` from a LOSC-format HDF file.

    Parameters
    ----------
    source : `str`, `list`, :class:`glue.lal.Cache`
        path to LOSC-format HDF5 file to read or cache of many files.
    channel : `str`
        name of HDF5 dataset to read.
    group : `str`, optional
        name of containing HDF5 group for ``channel``. If not given,
        the first dataset named ``channel`` will be assumed as the right
        one.
    start : `Time`, `~gwpy.time.LIGOTimeGPS`, optional
        start GPS time of desired data
    end : `Time`, `~gwpy.time.LIGOTimeGPS`, optional
        end GPS time of desired data


    Returns
    -------
    data : :class:`~gwpy.timeseries.statevector.StateVector`
        a new `TimeSeries` containing the data read from disk
    """
    kwargs.setdefault('target', StateVector)
    return read_losc_data_cache(*args, **kwargs)


@with_import('h5py')
def _find_dataset(h5group, name):
    """Find the named :class:`h5py.Dataset` in an HDF file.

    Parameters
    ----------
    h5group : :class:`h5py.File`, :class:`h5py.Group`
        open HDF file or group
    name : `str`
        name of :class:`h5py.Dataset` to find

    Returns
    -------
    data : :class:`h5ile.Dataset`
        HDF dataset
    """
    # find dataset directly
    if not isinstance(h5group, h5py.Group):
        raise ValueError("_find_dataset must be handed a h5py.Group object, "
                         "not %s" % h5group.__class__.__name__)
    if name in h5group and isinstance(h5group[name], h5py.Dataset):
        return h5group[name]
    # otherwise trawl through member groups
    for group in h5group.values():
        try:
            return _find_dataset(group, name)
        except ValueError:
            continue
    raise ValueError("Cannot find channel '%s' in file HDF object" % name)


registry.register_reader('losc', TimeSeries, read_losc_data_cache)
registry.register_reader('losc', StateVector, read_losc_state_cache)



# -*- coding: utf-8 -*-
# Copyright (C) Duncan Macleod (2013)
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""Array with metadata
"""

from __future__ import (division, print_function)

import os
import sys
import warnings
import re
from math import ceil

import numpy

from astropy import units

try:
    import nds2
except ImportError:
    NDS2_FETCH_TYPE_MASK = None
else:
    NDS2_FETCH_TYPE_MASK = (nds2.channel.CHANNEL_TYPE_RAW |
                            nds2.channel.CHANNEL_TYPE_RDS |
                            nds2.channel.CHANNEL_TYPE_TEST_POINT |
                            nds2.channel.CHANNEL_TYPE_STATIC)

from .. import version
from ..data import (Array2D, Series)
from ..detector import (Channel, ChannelList)
from ..io import (reader, datafind)
from ..time import (Time, to_gps)
from ..utils import (gprint, with_import)
from ..utils.docstring import interpolate_docstring
from ..utils.compat import OrderedDict

__author__ = 'Duncan Macleod <duncan.macleod@ligo.org>'
__version__ = version.version

__all__ = ['TimeSeriesBase', 'ArrayTimeSeries', 'TimeSeriesBaseDict']

_UFUNC_STRING = {'less': '<',
                 'less_equal': '<=',
                 'equal': '==',
                 'greater_equal': '>=',
                 'greater': '>',
                 }

interpolate_docstring.update({
    'time-axis': (
        """sample_rate : `float`, `~astropy.units.Quantity`, optional, default: `1`
        the rate of samples per second (Hertz)
    times : `array-like`
        the complete array of GPS times accompanying the data for this series.
        This argument takes precedence over `epoch` and `sample_rate` so should
        be given in place of these if relevant, not alongside"""),
})

@interpolate_docstring
class TimeSeriesBase(Series):
    """An `Array` with time-domain metadata.
    """
    _default_xunit = units.second
    _metadata_slots = ['name', 'channel', 'x0', 'dx']
    DictClass = None

    def __new__(cls, data, unit=None, times=None, epoch=None, channel=None,
                sample_rate=None, name=None, **kwargs):
        """Generate a new `TimeSeriesBase`.
        """
        # parse Channel input
        if isinstance(channel, Channel):
            name = name or channel.name
            unit = unit or channel.unit
            if sample_rate is None and times is None and 'dx' not in kwargs:
                sample_rate = channel.sample_rate
        if times is None and 'xindex' in kwargs:
            times = kwargs.pop('xindex')
        if sample_rate is None and times is None and 'dx' not in kwargs:
            sample_rate = 1
        if epoch is None and times is None and 'x0' not in kwargs:
            epoch = 0
        # generate TimeSeries
        new = super(TimeSeriesBase, cls).__new__(cls, data, name=name,
                                                 unit=unit, xindex=times,
                                                 channel=channel, **kwargs)
        if epoch is not None:
            new.epoch = epoch
        if sample_rate is not None:
            new.sample_rate = sample_rate
        return new

    # -------------------------------------------
    # TimeSeries properties

    dt = Series.dx
    span = Series.xspan

    @property
    def epoch(self):
        """GPS epoch for these data.

        This attribute is stored internally by the `x0` attribute

        :type: `~astropy.time.Time`
        """
        if self.x0 is None:
            return None
        else:
            return Time(self.x0, format='gps', scale='utc')

    @epoch.setter
    def epoch(self, epoch):
        if epoch is None:
            self.x0 = None
        elif isinstance(epoch, Time):
            self.x0 = epoch.gps
        else:
            try:
                self.x0 = to_gps(epoch)
            except TypeError:
                self.x0 = epoch

    @property
    def sample_rate(self):
        """Data rate for this `TimeSeries` in samples per second (Hertz).

        This attribute is stored internally by the `dx` attribute

        :type: `~astropy.units.Quantity` scalar
        """
        return (1 / self.dx).to('Hertz')

    @sample_rate.setter
    def sample_rate(self, val):
        if val is None:
            del self.dx
            return
        self.dx = (1 / units.Quantity(val, units.Hertz)).to(self.xunit)
        if numpy.isclose(self.dx.value, round(self.dx.value)):
            self.dx = units.Quantity(round(self.dx.value), self.dx.unit)

    @property
    def duration(self):
        """Duration of this series in seconds
        """
        return units.Quantity(self.span[1] - self.span[0], self.xunit)

    times = property(fget=Series.xindex.__get__,
                     fset=Series.xindex.__set__,
                     fdel=Series.xindex.__delete__,
                     doc="""Series of GPS times for each sample""")

    # -------------------------------------------
    # TimeSeries accessors

    interpolate_docstring.update({
        'timeseries-read1': """source : `str`, `~glue.lal.Cache`
            source of data, any of the following:

            - `str` path of single data file
            - `str` path of LAL-format cache file
            - `~glue.lal.Cache` describing one or more data files,

        channel : `str`, `~gwpy.detector.Channel`
            the name of the channel to read, or a `Channel` object.

        start : `~gwpy.time.LIGOTimeGPS`, `float`, `str`, optional
            GPS start time of required data, defaults to start of data found;
            any input parseable by `~gwpy.time.to_gps` is fine

        end : `~gwpy.time.LIGOTimeGPS`, `float`, `str`, optional
            GPS end time of required data, defaults to end of data found;
            any input parseable by `~gwpy.time.to_gps` is fine""",

        'timeseries-read2': """format : `str`, optional
            source format identifier. If not given, the format will be
            detected if possible. See below for list of acceptable
            formats.

        nproc : `int`, optional, default: `1`
            number of parallel processes to use, serial process by
            default.

            .. note::

               Parallel frame reading, via the ``nproc`` keyword argument,
               is only available when giving a `~glue.lal.Cache` of
               frames, or using the ``format='cache'`` keyword argument.

        gap : `str`, optional
            how to handle gaps in the cache, one of

            - 'ignore': do nothing, let the undelying reader method handle it
            - 'warn': do nothing except print a warning to the screen
            - 'raise': raise an exception upon finding a gap (default)
            - 'pad': insert a value to fill the gaps

        pad : `float`, optional
            value with which to fill gaps in the source data, only used if
            gap is not given, or `gap='pad'` is given""",

        'timeseries-fetch1': """channel : `str`, `~gwpy.detector.Channel`
            the name of the channel to read, or a `Channel` object.

        start : `~gwpy.time.LIGOTimeGPS`, `float`, `str`
            GPS start time of required data,
            any input parseable by `~gwpy.time.to_gps` is fine

        end : `~gwpy.time.LIGOTimeGPS`, `float`, `str`
            GPS end time of required data,
            any input parseable by `~gwpy.time.to_gps` is fine""",

        'timeseries-fetch2': """host : `str`, optional
            URL of NDS server to use, defaults to observatory site host

        port : `int`, optional
            port number for NDS server query, must be given with `host`

        verify : `bool`, optional, default: `True`
            check channels exist in database before asking for data

        connection : :class:`~gwpy.io.nds.NDS2Connection`
            open NDS connection to use

        verbose : `bool`, optional
            print verbose output about NDS progress

        type : `int`, optional
            NDS2 channel type integer

        dtype : `type`, `numpy.dtype`, `str`, optional
            identifier for desired output data type""",
    })

    @classmethod
    @interpolate_docstring
    @with_import('nds2')
    def fetch(cls, channel, start, end, host=None, port=None, verbose=False,
              connection=None, verify=False, pad=None,
              type=NDS2_FETCH_TYPE_MASK, dtype=None):
        """Fetch data from NDS into a `TimeSeries`.

        Parameters
        ----------
        %(timeseries-fetch1)s

        %(timeseries-fetch2)s
        """
        return cls.DictClass.fetch(
            [channel], start, end, host=host, port=port,
            verbose=verbose, connection=connection, verify=verify,
            pad=pad, type=type, dtype=dtype)[str(channel)]

    @classmethod
    @interpolate_docstring
    def find(cls, channel, start, end, frametype=None,
             pad=None, dtype=None, nproc=1, verbose=False, **readargs):
        """Find and read data from frames for a channel

        Parameters
        ----------
        %(timeseries-fetch1)s

        frametype : `str`, optional
            name of frametype in which this channel is stored, will search
            for containing frame types if necessary

        pad : `float`, optional
            value with which to fill gaps in the source data, only used if
            gap is not given, or `gap='pad'` is given

        nproc : `int`, optional, default: `1`
            number of parallel processes to use, serial process by
            default.

        dtype : `numpy.dtype`, `str`, `type`, or `dict`
            numeric data type for returned data, e.g. `numpy.float`, or
            `dict` of (`channel`, `dtype`) pairs

        verbose : `bool`, optional
            print verbose output about NDS progress.

        **readargs
            any other keyword arguments to be passed to `.read()`
        """
        return cls.DictClass.find(
            [channel], start, end, frametype=frametype, verbose=verbose,
            pad=pad, dtype=dtype, nproc=nproc, **readargs)[str(channel)]

    @classmethod
    @interpolate_docstring
    def get(cls, channel, start, end, pad=None, dtype=None, verbose=False,
            **kwargs):
        """Get data for this channel from frames or NDS

        This method dynamically accesses either frames on disk, or a
        remote NDS2 server to find and return data for the given interval

        Parameters
        ----------
        %(timeseries-fetch1)s

        pad : `float`, optional
            value with which to fill gaps in the source data, only used if
            gap is not given, or ``gap='pad'`` is given
        dtype : `numpy.dtype`, `str`, `type`, or `dict`
            numeric data type for returned data, e.g. `numpy.float`, or
            `dict` of (`channel`, `dtype`) pairs
        nproc : `int`, optional, default: `1`
            number of parallel processes to use, serial process by
            default.
        verbose : `bool`, optional
            print verbose output about NDS progress.
        **kwargs            other keyword arguments to pass to either
            :meth:`.find` (for direct GWF file access) or
            :meth:`.fetch` for remote NDS2 access

        See Also
        --------
        TimeSeries.fetch
            for grabbing data from a remote NDS2 server
        TimeSeries.find
            for discovering and reading data from local GWF files
        """
        return cls.DictClass.get(
            [channel], start, end, pad=pad, dtype=dtype, verbose=verbose,
            **kwargs)[str(channel)]

    # -------------------------------------------
    # Utilities

    def plot(self, **kwargs):
        """Plot the data for this `TimeSeries`
        """
        from ..plotter import TimeSeriesPlot
        return TimeSeriesPlot(self, **kwargs)

    @classmethod
    @with_import('nds2')
    def from_nds2_buffer(cls, buffer_, **metadata):
        """Construct a new `TimeSeries` from an `nds2.buffer` object

        Parameters
        ----------
        buffer_ : `nds2.buffer`
            the input NDS2-client buffer to read
        **metadata
            any other metadata keyword arguments to pass to the `TimeSeries`
            constructor

        Returns
        -------
        timeseries : `TimeSeries`
            a new `TimeSeries` containing the data from the `nds2.buffer`,
            and the appropriate metadata

        Notes
        -----
        This classmethod requires the nds2-client package
        """
        # cast as TimeSeries and return
        epoch = Time(buffer_.gps_seconds, buffer_.gps_nanoseconds,
                     format='gps')
        channel = Channel.from_nds2(buffer_.channel)
        return cls(buffer_.data, epoch=epoch, channel=channel, **metadata)

    @classmethod
    @with_import('lal')
    def from_lal(cls, lalts, copy=True):
        """Generate a new TimeSeries from a LAL TimeSeries of any type.
        """
        from ..utils.lal import from_lal_unit
        try:
            unit = from_lal_unit(lalts.sampleUnits)
        except TypeError:
            unit = None
        channel = Channel(lalts.name, sample_rate=1/lalts.deltaT, unit=unit,
                          dtype=lalts.data.data.dtype)
        out = cls(lalts.data.data, channel=channel, epoch=float(lalts.epoch),
                  copy=False)
        if copy:
            return out.copy()
        else:
            return out

    @with_import('lal')
    def to_lal(self):
        """Convert this `TimeSeries` into a LAL TimeSeries.
        """
        from ..utils.lal import (LAL_TYPE_STR_FROM_NUMPY, to_lal_unit)
        typestr = LAL_TYPE_STR_FROM_NUMPY[self.dtype.type]
        try:
            unit = to_lal_unit(self.unit)
        except (TypeError, AttributeError):
            try:
                unit = lal.DimensionlessUnit
            except AttributeError:
                unit = lal.lalDimensionlessUnit
        create = getattr(lal, 'Create%sTimeSeries' % typestr.upper())
        lalts = create(self.name, lal.LIGOTimeGPS(self.epoch.gps), 0,
                       self.dt.value, unit, self.size)
        lalts.data.data = self.value
        return lalts

    @classmethod
    def from_pycbc(cls, ts):
        """Convert a `pycbc.types.timeseries.TimeSeries` into a `TimeSeries`

        Parameters
        ----------
        ts : `pycbc.types.timeseries.TimeSeries`
            the input PyCBC `~pycbc.types.timeseries.TimeSeries` array

        Returns
        -------
        timeseries : `TimeSeries`
            a GWpy version of the input timeseries
        """
        return cls(ts.data, epoch=ts.epoch, sample_rate=1/ts.delta_t)

    @with_import('pycbc.types')
    def to_pycbc(self, copy=True):
        """Convert this `TimeSeries` into a PyCBC
        `~pycbc.types.timeseries.TimeSeries`

        Parameters
        ----------
        copy : `bool`, optional, default: `True`
            if `True`, copy these data to a new array

        Returns
        -------
        timeseries : `~pycbc.types.timeseries.TimeSeries`
            a PyCBC representation of this `TimeSeries`
        """
        return types.TimeSeries(self.data,
                                delta_t=self.dx.to('s').value,
                                epoch=self.epoch.gps, copy=copy)

    # -------------------------------------------
    # TimeSeries operations

    def __array_wrap__(self, obj, context=None):
        """Wrap an array into a TimeSeries, or a StateTimeSeries if
        dtype == bool.
        """
        if obj.dtype == numpy.dtype(bool):
            from .statevector import StateTimeSeries
            ufunc = context[0]
            value = context[1][-1]
            try:
                op_ = _UFUNC_STRING[ufunc.__name__]
            except KeyError:
                op_ = ufunc.__name__
            result = obj.view(StateTimeSeries)
            result.name = '%s %s %s' % (obj.name, op_, value)
            if hasattr(obj, 'unit') and str(obj.unit):
                result.name += ' %s' % str(obj.unit)
        else:
            result = super(TimeSeriesBase, self).__array_wrap__(
                obj, context=context)
        return result


class ArrayTimeSeries(TimeSeriesBase, Array2D):
    _default_xunit = TimeSeriesBase._default_xunit

    def __new__(cls, data, times=None, epoch=None, channel=None, unit=None,
                sample_rate=None, name=None, **kwargs):
        """Generate a new ArrayTimeSeries.
        """
        # parse Channel input
        if channel:
            channel = (isinstance(channel, Channel) and channel or
                       Channel(channel))
            name = name or channel.name
            unit = unit or channel.unit
            sample_rate = sample_rate or channel.sample_rate
        # generate TimeSeries
        new = Array2D.__new__(cls, data, name=name, unit=unit, epoch=epoch,
                              channel=channel, sample_rate=sample_rate,
                              times=times, **kwargs)
        return new


def as_series_dict_class(seriesclass):
    """Decorate a `dict` class to declare itself as the `DictClass` for
    its `EntryClass`

    This method should be used to decorate sub-classes of the
    `TimeSeriesBaseDict` to provide a reference to that class from the
    relevant subclass of `TimeSeriesBase`.
    """
    def decorate_class(cls):
        seriesclass.DictClass = cls
        return cls
    return decorate_class


@as_series_dict_class(TimeSeriesBase)
class TimeSeriesBaseDict(OrderedDict):
    """Ordered key-value mapping of named `TimeSeriesBase` objects

    This object is designed to hold data for many different sources (channels)
    for a single time span.

    The main entry points for this object are the
    :meth:`~TimeSeriesBaseDict.read` and :meth:`~TimeSeriesBaseDict.fetch`
    data access methods.
    """
    EntryClass = TimeSeriesBase

    # use input/output registry to allow multi-format reading
    read = classmethod(reader(doc="""
        Read data into a `TimeSeriesBaseDict`.

        Parameters
        ----------
        source : `str`, `~glue.lal.Cache`
            a single file path `str`, or a `~glue.lal.Cache` containing
            a contiguous list of files.
        channels : `~gwpy.detector.channel.ChannelList`, `list`
            a list of channels to read from the source.
        start : `~gwpy.time.Time`, `float`, optional
            GPS start time of required data.
        end : `~gwpy.time.Time`, `float`, optional
            GPS end time of required data.
        format : `str`, optional
            source format identifier. If not given, the format will be
            detected if possible. See below for list of acceptable
            formats.
        nproc : `int`, optional, default: ``1``
            number of parallel processes to use, serial process by
            default.

            .. note::

               Parallel frame reading, via the ``nproc`` keyword argument,
               is only available when giving a :class:`~glue.lal.Cache` of
               frames, or using the ``format='cache'`` keyword argument.

        gap : `str`, optional
            how to handle gaps in the cache, one of

            - 'ignore': do nothing, let the undelying reader method handle it
            - 'warn': do nothing except print a warning to the screen
            - 'raise': raise an exception upon finding a gap (default)
            - 'pad': insert a value to fill the gaps

        pad : `float`, optional
            value with which to fill gaps in the source data, only used if
            gap is not given, or `gap='pad'` is given

        Returns
        -------
        dict : `TimeSeriesBaseDict`
            a new `TimeSeriesBaseDict` containing data for the given channel.

        Raises
        ------
        Exception
            if no format could be automatically identified.

        Notes
        -----"""))

    def __iadd__(self, other):
        return self.append(other)

    def copy(self):
        new = self.__class__()
        for key, val in self.iteritems():
            new[key] = val.copy()
        return new

    def append(self, other, copy=True, **kwargs):
        for key, ts in other.iteritems():
            if key in self:
                self[key].append(ts, **kwargs)
            elif copy:
                self[key] = ts.copy()
            else:
                self[key] = ts
        return self

    def prepend(self, other, **kwargs):
        for key, ts in other.iteritems():
            if key in self:
                self[key].prepend(ts, **kwargs)
            else:
                self[key] = ts
        return self

    def crop(self, start=None, end=None, copy=False):
        """Crop each entry of this `TimeSeriesBaseDict`.

        This method calls the :meth:`crop` method of all entries and
        modifies this dict in place.

        Parameters
        ----------
        start : `Time`, `float`
            GPS start time to crop `TimeSeries` at left
        end : `Time`, `float`
            GPS end time to crop `TimeSeries` at right

        See Also
        --------
        TimeSeries.crop
            for more details
        """
        for key, val in self.iteritems():
            self[key] = val.crop(start=start, end=end, copy=copy)
        return self

    def resample(self, rate, **kwargs):
        """Resample items in this dict.

        This operation over-writes items inplace.

        Parameters
        ----------
        rate : `dict`, `float`
            either a `dict` of (channel, `float`) pairs for key-wise
            resampling, or a single float/int to resample all items.
        kwargs
             other keyword arguments to pass to each item's resampling
             method.
        """
        if not isinstance(rate, dict):
            rate = dict((c, rate) for c in self)
        for key, resamp in rate.iteritems():
            self[key] = self[key].resample(resamp, **kwargs)
        return self

    @classmethod
    @with_import('nds2')
    def fetch(cls, channels, start, end, host=None, port=None,
              verify=False, verbose=False, connection=None,
              pad=None, type=NDS2_FETCH_TYPE_MASK, dtype=None):
        """Fetch data from NDS for a number of channels.

        Parameters
        ----------
        channels : `list`
            required data channels.
        start : `~gwpy.time.Time`, or float
            GPS start time of data span.
        end : `~gwpy.time.Time`, or float
            GPS end time of data span.
        host : `str`, optional
            URL of NDS server to use, defaults to observatory site host.
        port : `int`, optional
            port number for NDS server query, must be given with `host`.
        verify : `bool`, optional, default: `True`
            check channels exist in database before asking for data
        verbose : `bool`, optional
            print verbose output about NDS progress.
        connection : :class:`~gwpy.io.nds.NDS2Connection`
            open NDS connection to use.
        type : `int`, `str`,
            NDS2 channel type integer or string name.
        dtype : `numpy.dtype`, `str`, `type`, or `dict`
            numeric data type for returned data, e.g. `numpy.float`, or
            `dict` of (`channel`, `dtype`) pairs

        Returns
        -------
        data : :class:`~gwpy.timeseries.TimeSeriesBaseDict`
            a new `TimeSeriesBaseDict` of (`str`, `TimeSeries`) pairs fetched
            from NDS.
        """
        from ..segments import (Segment, SegmentList)
        from ..io import nds as ndsio
        # parse times
        start = to_gps(start)
        end = to_gps(end)
        istart = start.seconds
        iend = ceil(end)

        # test S6 h(t) channel
        if any(re.match('[HL]1:LDAS-STRAIN\Z', str(c)) for c in channels):
            verify = True

        # parse dtype
        if isinstance(dtype, (tuple, list)):
            dtype = [numpy.dtype(r) if r is not None else None for r in dtype]
            dtype = dict(zip(channels, dtype))
        elif not isinstance(dtype, dict):
            if dtype is not None:
                dtype = numpy.dtype(dtype)
            dtype = dict((channel, dtype) for channel in channels)

        # open connection for specific host
        if host and not port and re.match('[a-z]1nds[0-9]\Z', host):
            port = 8088
        elif host and not port:
            port = 31200
        if host is not None and port is not None and connection is None:
            if verbose:
                gprint("Connecting to %s:%s..." % (host, port), end=' ')
            connection = ndsio.auth_connect(host, port)
            if verbose:
                gprint("Connected.")
        elif connection is not None and verbose:
            gprint("Received connection to %s:%d."
                   % (connection.get_host(), connection.get_port()))
        # otherwise cycle through connections in logical order
        if connection is None:
            ifos = set([Channel(channel).ifo for channel in channels])
            if len(ifos) == 1:
                ifo = list(ifos)[0]
            else:
                ifo = None
            hostlist = ndsio.host_resolution_order(ifo, epoch=start)
            for host, port in hostlist:
                try:
                    return cls.fetch(channels, start, end, host=host,
                                     port=port, verbose=verbose, type=type,
                                     verify=verify, dtype=dtype, pad=pad)
                except (RuntimeError, ValueError) as e:
                    if verbose:
                        gprint('Something went wrong:', file=sys.stderr)
                        # if error and user supplied their own server, raise
                        warnings.warn(str(e), ndsio.NDSWarning)

            # if we got this far, we can't get all of the channels in one go
            if len(channels) > 1:
                return cls(
                    (c, cls.EntryClass.fetch(c, start, end, verbose=verbose,
                                             type=type, verify=verify,
                                             dtype=dtype.get(c), pad=pad))
                    for c in channels)
            e = "Cannot find all relevant data on any known server."
            if not verbose:
                e += (" Try again using the verbose=True keyword argument to "
                      "see detailed failures.")
            raise RuntimeError(e)

        # at this point we must have an open connection, so we can proceed
        # normally

        # verify channels
        if verify:
            if verbose:
                gprint("Checking channels against the NDS database...",
                       end=' ')
            else:
                warnings.filterwarnings('ignore', category=ndsio.NDSWarning,
                                        append=False)
            try:
                qchannels = ChannelList.query_nds2(channels,
                                                   connection=connection,
                                                   type=type, unique=True)
            except ValueError as e:
                try:
                    channels2 = ['%s*' % c for c in map(str, channels)]
                    qchannels = ChannelList.query_nds2(channels2,
                                                       connection=connection,
                                                       type=type, unique=True)
                except ValueError:
                    raise e
            if verbose:
                gprint("Complete.")
            else:
                warnings.filters.pop(0)
        else:
            qchannels = ChannelList(map(Channel, channels))

        # test for minute trends
        if (any([c.type == 'm-trend' for c in qchannels]) and
                (start % 60 or end % 60)):
            warnings.warn("Requested at least one minute trend, but "
                          "start and stop GPS times are not modulo "
                          "60-seconds (from GPS epoch). Times will be "
                          "expanded outwards to compensate")
            if start % 60:
                start = int(start) // 60 * 60
                istart = start
            if end % 60:
                end = int(end) // 60 * 60 + 60
                iend = end
            have_minute_trends = True
        else:
            have_minute_trends = False

        # get segments for data
        allsegs = SegmentList([Segment(istart, iend)])
        qsegs = SegmentList([Segment(istart, iend)])
        if pad is not None:
            from subprocess import CalledProcessError
            try:
                segs = ChannelList.query_nds2_availability(
                    channels, istart, iend, host=connection.get_host())
            except (RuntimeError, CalledProcessError) as e:
                warnings.warn(str(e), ndsio.NDSWarning)
            else:
                for channel in segs:
                    try:
                        csegs = sorted(segs[channel].values(),
                                       key=lambda x: abs(x))[-1]
                    except IndexError:
                        csegs = SegmentList([])
                    qsegs &= csegs

            if verbose:
                gprint('Found %d viable segments of data with %.2f%% coverage'
                       % (len(qsegs), abs(qsegs) / abs(allsegs) * 100))

        out = cls()
        for (istart, iend) in qsegs:
            istart = int(istart)
            iend = int(iend)
            # fetch data
            if verbose:
                gprint('Downloading data... ', end='\r')

            # determine buffer duration
            data = connection.iterate(istart, iend,
                                      [c.ndsname for c in qchannels])
            nsteps = 0
            i = 0
            for buffers in data:
                for buffer_, c in zip(buffers, channels):
                    ts = cls.EntryClass.from_nds2_buffer(
                        buffer_, dtype=dtype.get(c))
                    out.append({c: ts}, pad=pad,
                               gap=pad is None and 'raise' or 'pad')
                if not nsteps:
                    if have_minute_trends:
                        dur = buffer_.length * 60
                    else:
                        dur = buffer_.length / buffer_.channel.sample_rate
                    nsteps = ceil((iend - istart) / dur)
                i += 1
                if verbose:
                    gprint('Downloading data... %d%%' % (100 * i // nsteps),
                           end='\r')
                    if i == nsteps:
                        gprint('')

        # pad to end of request if required
        if len(qsegs) and iend < float(end):
            dt = float(end) - float(iend)
            for channel in out:
                nsamp = dt * out[channel].sample_rate.value
                out[channel].append(
                    numpy.ones(nsamp, dtype=out[channel].dtype) * pad)
        # match request exactly
        for channel in out:
            if istart > start or iend < end:
                out[channel] = out[channel].crop(start, end)

        if verbose:
            gprint('Success.')
        return out

    @classmethod
    def find(cls, channels, start, end, frametype=None,
             pad=None, dtype=None, nproc=1, verbose=False,
             allow_tape=True, **readargs):
        """Find and read data from frames for a number of channels.

        Parameters
        ----------
        channels : `list`
            required data channels.
        start : `~gwpy.time.Time`, or float
            GPS start time of data span.
        end : `~gwpy.time.Time`, or float
            GPS end time of data span.
        frametype : `str`, optional
            name of frametype in which this channel is stored, by default
            will search for all required frame types
        pad : `float`, optional
            value with which to fill gaps in the source data, only used if
            gap is not given, or `gap='pad'` is given
        dtype : `numpy.dtype`, `str`, `type`, or `dict`
            numeric data type for returned data, e.g. `numpy.float`, or
            `dict` of (`channel`, `dtype`) pairs
        nproc : `int`, optional, default: `1`
            number of parallel processes to use, serial process by
            default.
        verbose : `bool`, optional
            print verbose output about NDS progress.
        allow_tape : `bool`, optional, default: `True`
            allow reading from frames on tape
        **readargs
            any other keyword arguments to be passed to `.read()`
        """
        start = to_gps(start)
        end = to_gps(end)
        # -- find frametype(s)
        if frametype is None:
            frametypes = dict()
            for c in channels:
                ft = datafind.find_best_frametype(c, start, end,
                                                  allow_tape=allow_tape)
                try:
                    frametypes[ft].append(c)
                except KeyError:
                    frametypes[ft] = [c]
            if verbose and len(frametypes) > 1:
                gprint("Determined %d frametypes to read" % len(frametypes))
            elif verbose:
                gprint("Determined best frametype as %r"
                       % frametypes.keys()[0])
        else:
            frametypes = {frametype: channels}
        # -- read data
        out = cls()
        for ft, clist in frametypes.iteritems():
            if verbose:
                gprint("Reading data from %s frames..." % ft, end=' ')
            # find frames
            channellist = ChannelList.from_names(*clist)
            ifos = ''.join(sorted(set(c.ifo[0] for c in channellist)))
            connection = datafind.connect()
            cache = connection.find_frame_urls(ifos, ft, start, end,
                                               urltype='file')
            # read data
            readargs.setdefault('format', 'gwf')
            out.append(cls.read(cache, clist, start=start, end=end, pad=pad,
                                dtype=dtype, nproc=nproc, **readargs))
            if verbose:
                gprint("Done")
        return out

    @classmethod
    def get(cls, channels, start, end, pad=None, dtype=None, verbose=False,
            allow_tape=False, **kwargs):
        """Retrieve data for multiple channels from frames or NDS

        This method dynamically accesses either frames on disk, or a
        remote NDS2 server to find and return data for the given interval

        Parameters
        ----------
        channels : `list`
            required data channels.
        start : `~gwpy.time.Time`, or float
            GPS start time of data span.
        end : `~gwpy.time.Time`, or float
            GPS end time of data span.
        frametype : `str`, optional
            name of frametype in which this channel is stored, by default
            will search for all required frame types
        pad : `float`, optional
            value with which to fill gaps in the source data, only used if
            gap is not given, or `gap='pad'` is given
        dtype : `numpy.dtype`, `str`, `type`, or `dict`
            numeric data type for returned data, e.g. `numpy.float`, or
            `dict` of (`channel`, `dtype`) pairs
        nproc : `int`, optional, default: `1`
            number of parallel processes to use, serial process by
            default.
        verbose : `bool`, optional
            print verbose output about NDS progress.
        allow_tape : `bool`, optional, default: `False`
            allow the use of frames that are held on tape, default is `False`
            to attempt to allow the `TimeSeries.fetch` method to
            intelligently select a server that doesn't use tapes for
            data storage (doesn't always work)
        **kwargs
            other keyword arguments to pass to either
            `TimeSeriesBaseDict.find` (for direct GWF file access) or
            `TimeSeriesBaseDict.fetch` for remote NDS2 access
        """
        try_frames = True
        # work out whether to use NDS2 or frames
        if not os.getenv('LIGO_DATAFIND_SERVER'):
            try_frames = False
        host = kwargs.get('host', None)
        if host is not None and host.startswith('nds'):
            try_frames = False
        # try and find from frames
        if try_frames:
            if verbose:
                gprint("Attempting to access data from frames...")
            try:
                return cls.find(channels, start, end, pad=pad, dtype=dtype,
                                verbose=verbose, allow_tape=allow_tape,
                                **kwargs)
            except (RuntimeError, ValueError) as e:
                if verbose:
                    gprint(str(e), file=sys.stderr)
                    gprint("Failed to access data from frames, trying NDS...")
        # otherwise fetch from NDS
        kwargs.pop('nproc', None)
        kwargs.pop('frametype', None)
        try:
            return cls.fetch(channels, start, end, pad=pad, dtype=dtype,
                             verbose=verbose, **kwargs)
        except RuntimeError as e:
            # if all else fails, try and get each channel individually
            if len(channels) == 1:
                raise
            else:
                if verbose:
                    gprint(str(e), file=sys.stderr)
                    gprint("Failed to access data for all channels as a "
                           "group, trying individually:")
                return cls(
                    (c, cls.EntryClass.get(c, start, end, pad=pad, dtype=dtype,
                                           verbose=verbose, **kwargs))
                    for c in channels)

    def plot(self, label='key', **kwargs):
        """Plot the data for this `TimeSeriesBaseDict`.

        Parameters
        ----------
        label : `str`, optional
            labelling system to use, or fixed label for all elements
            Special values include

            - ``'key'``: use the key of the `TimeSeriesBaseDict`,
            - ``'name'``: use the :attr:`~TimeSeries.name` of each element

            If anything else, that fixed label will be used for all lines.

        **kwargs
            all other keyword arguments are passed to the plotter as
            appropriate
        """
        from ..plotter import TimeSeriesPlot
        figargs = dict()
        for key in ['figsize', 'dpi']:
            if key in kwargs:
                figargs[key] = kwargs.pop(key)
        plot_ = TimeSeriesPlot(**figargs)
        ax = plot_.gca()
        for lab, ts in self.iteritems():
            if label.lower() == 'name':
                lab = ts.name
            elif label.lower() != 'key':
                lab = label
            ax.plot(ts, label=lab, **kwargs)
        return plot_


class TimeSeriesBaseList(list):
    """Fancy list representing a list of `TimeSeriesBase`

    The `TimeSeriesBaseList` provides an easy way to collect and organise
    `TimeSeriesBase` for a single `Channel` over multiple segments.

    Parameters
    ----------
    *items
        any number of `TimeSeriesBase`

    Returns
    -------
    list
        a new `TimeSeriesBaseList`

    Raises
    ------
    TypeError
        if any elements are not `TimeSeriesBase`
    """
    EntryClass = TimeSeriesBase

    def __init__(self, *items):
        """Initialise a new list
        """
        super(TimeSeriesBaseList, self).__init__()
        for item in items:
            self.append(item)

    @property
    def segments(self):
        from ..segments import SegmentList
        return SegmentList([item.span for item in self])

    def append(self, item):
        if not isinstance(item, self.EntryClass):
            raise TypeError("Cannot append type '%s' to %s"
                            % (item.__class__.__name__,
                               self.__class__.__name__))
        super(TimeSeriesBaseList, self).append(item)
        return self
    append.__doc__ = list.append.__doc__

    def extend(self, item):
        item = TimeSeriesBaseList(item)
        super(TimeSeriesBaseList, self).extend(item)
    extend.__doc__ = list.extend.__doc__

    def coalesce(self):
        """Merge contiguous elements of this list into single objects

        This method implicitly sorts and potentially shortens the this list.
        """
        self.sort(key=lambda ts: ts.x0.value)
        i = j = 0
        N = len(self)
        while j < N:
            this = self[j]
            j += 1
            if j < N and this.is_contiguous(self[j]) == 1:
                while j < N and this.is_contiguous(self[j]):
                    try:
                        this = self[i] = this.append(self[j])
                    except ValueError as e:
                        if 'cannot resize this array' in str(e):
                            this = this.copy()
                            this = self[i] = this.append(self[j])
                        else:
                            raise
                    j += 1
            else:
                self[i] = this
            i += 1
        del self[i:]
        return self

    def join(self, pad=0.0, gap='raise'):
        """Concatenate all of the elements of this list into a single object

        Parameters
        ----------
        pad : `float`, optional, default: `0.0`
            value with which to pad gaps
        gap : `str`, optional, default: `'raise'`
            what to do in the event of a discontguity in the data, one of

            - 'raise': raise an exception
            - 'warn': print a warning
            - 'ignore': append series as if there was no gap
            - 'pad': pad the gap with a value

        Returns
        -------
        `TimeSeriesBase`
             a single `TimeSeriesBase` covering the full span of all entries
             in this list

        See Also
        --------
        TimeSeriesBase.append
            for details on how the individual series are concatenated together
        """
        if len(self) == 0:
            return self.EntryClass(numpy.empty((0,) * self.EntryClass._ndim))
        self.sort(key=lambda t: t.epoch.gps)
        out = self[0].copy()
        for ts in self[1:]:
            out.append(ts, gap=gap, pad=pad)
        return out



from gwpy.timeseries import TimeSeries
gwdata = TimeSeries.fetch('H1:LDAS-STRAIN', 'September 16 2010 06:40',
                          'September 16 2010 06:50')
specgram = gwdata.spectrogram(20, fftlength=8, overlap=4) ** (1/2.)
plot = specgram.plot(norm='log', vmin=1e-23, vmax=1e-19)
ax = plot.gca()
ax.set_ylim(40, 4000)
ax.set_yscale('log')
plot.add_colorbar(label='Gravitational-wave strain [m/$\sqrt{\mathrm{Hz}}$]')
plot.show()



# -*- coding: utf-8 -*-
# Copyright (C) Duncan Macleod (2013)
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""Unit test for astro module
"""

import os
import os.path
import tempfile

from compat import unittest

from astropy import units

from gwpy import (astro, version)
from gwpy.timeseries import TimeSeries

__author__ = 'Duncan Macleod <duncan.macleod@ligo.org>'
__version__ = version.version


class AstroTests(unittest.TestCase):
    """`TestCase` for the astro module
    """
    framefile = os.path.join(os.path.split(__file__)[0], 'data',
                             'HLV-GW100916-968654552-1.hdf')
    tmpfile = '%s.%%s' % tempfile.mktemp(prefix='gwpy_test_')

    def setUp(self):
        # read data
        self.data = TimeSeries.read(self.framefile, 'L1:LDAS-STRAIN')
        # calculate PSD
        self.psd = self.data.psd(0.4, 0.2, window=('kaiser', 24))

    def test_inspiral_range(self):
        r = astro.inspiral_range(self.psd, fmin=40)
        self.assertEqual(r.unit, units.Mpc)
        self.assertAlmostEqual(r.value, 19.63704209223392)
        return r

    def test_inspiral_range_psd(self):
        r = astro.inspiral_range_psd(self.psd)
        self.assertEqual(r.unit, units.Mpc ** 2 / units.Hertz)
        self.assertAlmostEqual(r.max().value, 7.915847068684727)
        return r

    def test_burst_range(self):
        r = astro.burst_range(self.psd[self.psd.frequencies.value < 1000])
        self.assertEqual(r.unit, units.Mpc)
        self.assertAlmostEqual(r.value, 13.813232309724613)
        return r

    def test_burst_range_spectrum(self):
        r = astro.burst_range_spectrum(
            self.psd[self.psd.frequencies.value < 1000])
        self.assertEqual(r.unit, units.Mpc)
        self.assertAlmostEqual(r.max().value, 35.19303454822539)
        return r


if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-
# Copyright (C) Duncan Macleod (2013)
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""Unit test for `io` module
"""

import os
import tempfile

from compat import unittest

from gwpy import version
from gwpy.io.cache import (Cache, CacheEntry, cache_segments)
from gwpy.segments import (Segment, SegmentList)

__author__ = 'Duncan Macleod <duncan.macleod@ligo.org>'
__version__ = version.version


class IoTests(unittest.TestCase):

    def test_nds2_host_order_none(self):
        """Test `host_resolution_order` with `None` IFO
        """
        try:
            from gwpy.io import nds
        except ImportError as e:
            self.skipTest(str(e))
        hro = nds.host_resolution_order(None, env=None)
        self.assertListEqual(hro, [('nds.ligo.caltech.edu', 31200)])

    def test_nds2_host_order_ifo(self):
        """Test `host_resolution_order` with `ifo` argument
        """
        try:
            from gwpy.io import nds
        except ImportError as e:
            self.skipTest(str(e))
        hro = nds.host_resolution_order('L1', env=None)
        self.assertListEqual(
            hro, [('nds.ligo-la.caltech.edu', 31200),
                  ('nds.ligo.caltech.edu', 31200)])

    def test_nds2_host_order_ndsserver(self):
        """Test `host_resolution_order` with default env set
        """
        try:
            from gwpy.io import nds
        except ImportError as e:
            self.skipTest(str(e))
        os.environ['NDSSERVER'] = 'test1.ligo.org:80,test2.ligo.org:43'
        hro = nds.host_resolution_order(None)
        self.assertListEqual(
            hro, [('test1.ligo.org', 80), ('test2.ligo.org', 43),
                  ('nds.ligo.caltech.edu', 31200)])
        hro = nds.host_resolution_order('L1')
        self.assertListEqual(
            hro, [('test1.ligo.org', 80), ('test2.ligo.org', 43),
                  ('nds.ligo-la.caltech.edu', 31200),
                  ('nds.ligo.caltech.edu', 31200)])

    def test_nds2_host_order_env(self):
        """Test `host_resolution_order` with non-default env set
        """
        try:
            from gwpy.io import nds
        except ImportError as e:
            self.skipTest(str(e))
        os.environ['TESTENV'] = 'test1.ligo.org:80,test2.ligo.org:43'
        hro = nds.host_resolution_order(None, env='TESTENV')
        self.assertListEqual(
            hro, [('test1.ligo.org', 80), ('test2.ligo.org', 43),
                  ('nds.ligo.caltech.edu', 31200)])

    def test_nds2_host_order_epoch(self):
        """Test `host_resolution_order` with old GPS epoch
        """
        try:
            from gwpy.io import nds
        except ImportError as e:
            self.skipTest(str(e))
        # test kwarg doesn't change anything
        hro = nds.host_resolution_order('L1', epoch='now', env=None)
        self.assertListEqual(
            hro, [('nds.ligo-la.caltech.edu', 31200),
                  ('nds.ligo.caltech.edu', 31200)])
        # test old epoch puts CIT ahead of LLO
        hro = nds.host_resolution_order('L1', epoch='Jan 1 2015', env=None)
        self.assertListEqual(
            hro, [('nds.ligo.caltech.edu', 31200),
                  ('nds.ligo-la.caltech.edu', 31200)])
        # test epoch doesn't operate with env
        os.environ['TESTENV'] = 'test1.ligo.org:80,test2.ligo.org:43'
        hro = nds.host_resolution_order('L1', epoch='now', env='TESTENV')
        self.assertListEqual(
            hro, [('test1.ligo.org', 80), ('test2.ligo.org', 43),
                  ('nds.ligo-la.caltech.edu', 31200),
                  ('nds.ligo.caltech.edu', 31200)])

    @staticmethod
    def make_cache():
        segs = SegmentList()
        cache = Cache()
        for seg in [(0, 1), (1, 2), (4, 5)]:
            d = seg[1] - seg[0]
            _, f = tempfile.mkstemp(prefix='A-',
                                    suffix='-%d-%d.tmp' % (seg[0], d))
            cache.append(CacheEntry.from_T050017(f))
            segs.append(Segment(*seg))
        return cache, segs

    @staticmethod
    def destroy_cache(cache):
        for f in cache.pfnlist():
            if os.path.isfile(f):
                os.remove(f)

    def test_cache_segments(self):
        # check empty input
        sl = cache_segments()
        self.assertIsInstance(sl, SegmentList)
        self.assertEquals(len(sl), 0)
        cache, segs = self.make_cache()
        try:
            # check good cache
            sl = cache_segments(cache)
            self.assertNotEquals(sl, segs)
            self.assertEquals(sl, type(segs)(segs).coalesce())
            # check bad cache
            os.remove(cache[0].path)
            sl = cache_segments(cache)
            self.assertEquals(sl, segs[1:])
            # check cache with no existing files
            sl = cache_segments(cache[:1])
            self.assertEquals(sl, SegmentList())
            # check errors
            self.assertRaises(TypeError, cache_segments, blah='blah')
            self.assertRaises(ValueError, cache_segments, cache,
                              on_missing='error')
            self.assertRaises(ValueError, cache_segments, cache,
                             on_missing='blah')
        # clean up
        finally:
            self.destroy_cache(cache)


if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-
# Copyright (C) Duncan Macleod (2013)
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""Construct HTML maps on top of images.
"""

from itertools import izip

import numpy

from matplotlib.collections import Collection
from matplotlib.figure import Figure

from .. import version
from ..data import Series

__author__ = 'Duncan Macleod <duncan.macleod@ligo.org>'
__version__ = version.version

JQUERY_URL = '//ajax.googleapis.com/ajax/libs/jquery/1.11.1/jquery.min.js'

HTML_HEADER = """<!doctype html>
<html lang="en">
<head>
    <title>{title}</title>
    <script src="{jquery}"></script>
</head>
<body>
"""
HTML_FOOTER = "</body>\n</html>"

HTML_MAP = """
<img src="{png}" usemap="#{map}" width="{width}" height="{height}"/>
<map name="{map}">
    {data}
</map>
<script>
    var id_ = 'tmptooltip';
    $('area').mouseenter(function() {{
        var position = $(this).attr('coords').split(',');
        x = +position[0];
        y = +position[1];
        var $popup = $('<div />').appendTo('body');
        $popup.attr('id', id_);
        $popup.addClass('gwpy-tooltip');
        $popup.html('<p>'+$(this).attr('alt')+'</p>');
        $popup.css({{position: 'absolute', top: y, left: x+20}});
    }});
    $('area').mouseleave(function() {{
        $('#'+id_).remove();
    }});
</script>
"""


def html_area(x, y, href='#', alt=None, shape='circle', **kwargs):
    """Format the HTML <area /> tag for this (x, y) <map> element

    Parameters
    ----------
    x : `int`
        x pixel coordinate
    y : `int`
        y pixel coordinate
    href : `str`, optional
        URL for onclick
    alt : `str`, optional
        hover popup information, defaults to '(x, y)'
    shape : `str`, optional
        shape of <area />
    **kwargs
        any other attr="value" pairs to append to <area /> tag

    Returns
    -------
    area : `str`
        formatted line of HTML defining <area />
    """
    if alt is None:
        alt = '(%s, %s)' % (x, y)
    out = ('<area shape="{shape}" coords="{x:d},{y:d},5" href="{href}" '
           'alt="{alt}" '.format(shape=shape, x=x, y=y, href=href, alt=alt))
    for attr, value in kwargs.iteritems():
        out += '%s="%s" ' % (attr, value)
    out += '/>'
    return out


def map_artist(artist, filename, mapname='points', shape='circle',
               popup=None, title=None, standalone=True, jquery=JQUERY_URL):
    """Construct an HTML <map> to annotate the given `artist`

    Parameters
    ----------
    artist : `~matplotlib.artist.Artist`
        the plotted object (as returned from
        :meth:`~matplotlib.axes.Axes.plot`, for example)
    filename : `str`
        path to image file on disk
    mapname : `str`, optional
        ID to connect <img> tag and <map> tags, default: ``'points'``. This
        should be unique if multiple maps are to be written to a single HTML
        file.
    shape : `str`, optional
        shape for <area> tag, default: ``'circle'``
    popup : `function`, `iterable`, optional
        content for alt tooltip popup, either a function that can be
        called with the (x, y) pixel coords of each data point, or an
        iterable with one `str` text per data point
    standalone : `bool`, optional
        wrap map HTML with required HTML5 header and footer tags,
        default: `True`
    title : `str`, optional
        title name for standalone HTML page
    jquery : `str`, optional
        URL of jquery script

    Returns
    -------
    html : `str`
        formatted HTML page

    Notes
    -----
    This method should be run *after* the top-level figure is saved to disk,
    so that the relevant DPI and figure size information is fixed.
    """
    axes = artist.axes

    # get artist data
    if isinstance(artist, Collection):
        data = artist.get_offsets()
        z = artist.get_array()
        if z is not None:
            data = numpy.vstack((data[:, 0], data[:, 1], z)).T
    else:
        data = artist.get_data()

    return _map(data, axes, filename, mapname=mapname, shape=shape,
                popup=popup, title=title, standalone=standalone, jquery=jquery)


def map_data(data, axes, filename, mapname='points', shape='circle',
             popup=None, title=None, standalone=True, jquery=JQUERY_URL):
    """Construct an HTML <map> to annotate the given `artist`

    Parameters
    ----------
    data : `~gwpy.data.Series`, `~numpy.ndarray`
        data to map
    axes : `~matplotlib.axes.Axes`
        `Axes` to map onto
    filename : `str`
        path to image file on disk
    mapname : `str`, optional
        ID to connect <img> tag and <map> tags, default: ``'points'``. This
        should be unique if multiple maps are to be written to a single HTML
        file.
    shape : `str`, optional
        shape for <area> tag, default: ``'circle'``
    popup : `function`, `iterable`, optional
        content for alt tooltip popup, either a function that can be
        called with the (x, y) pixel coords of each data point, or an
        iterable with one `str` text per data point
    standalone : `bool`, optional
        wrap map HTML with required HTML5 header and footer tags,
        default: `True`
    title : `str`, optional
        title name for standalone HTML page
    jquery : `str`, optional
        URL of jquery script

    Returns
    -------
    html : `str`
        formatted HTML page

    Notes
    -----
    This method should be run *after* the top-level figure is saved to disk,
    so that the relevant DPI and figure size information is fixed.
    """
    if isinstance(data, Series):
        data = numpy.vstack((data.xindex.value, data.value)).T
    elif isinstance(data, numpy.ndarray):
        pass
    else:
        raise NotImplementedError('Custom mappings for %r types have not been '
                                  'implemented. Please plot these data and '
                                  'pass the artist to map_artist().')
    if isinstance(axes, Figure):
        axes = axes.gca()

    return _map(data, axes, filename, mapname=mapname, title=title,
                shape=shape, standalone=standalone, jquery=jquery)


def _map(data, axes, filename, href='#', mapname='points', popup=None,
         title='', shape='circle', standalone=True, jquery=JQUERY_URL):
    """Build an HTML <map> for a data set on some axes.
    """
    fig = axes.figure
    transform = axes.transData

    # get 2-d pixels
    pixels = transform.transform(data[:, :2]).astype(int)

    # get figure size
    dpi = fig.dpi
    width = int(fig.get_figwidth() * dpi)
    height = int(fig.get_figheight() * dpi)

    npix = pixels.shape[0]
    # configure hrefs
    if isinstance(href, str):
        hrefs = [href] * npix
    elif callable(href):
        hrefs = map(href, data)
    else:
        hrefs = numpy.asarray(href)
    if not len(hrefs) == npix:
        raise ValueError("href map with %d elements doesn't match %d data "
                         "points" % (hrefs.shape[0], npix))

    # build map
    areas = []
    for i, (datum, pixel, href) in enumerate(
            izip(data[::-1], pixels[::-1], hrefs[::-1])):
        if callable(popup):
            alt = popup(*datum)
        elif popup is None:
            alt = '(%s)' % (', '.join(map(str, datum)))
        else:
            alt = popup[-i-1]
        x = pixel[0]
        y = height - pixel[1]
        areas.append(html_area(x, y, href=href, alt=alt, shape=shape))

    hmap = HTML_MAP.format(title=title, png=filename, map=mapname, width=width,
                           height=height, data='\n    '.join(areas))
    if standalone:
        return (HTML_HEADER.format(title=title, jquery=jquery) + hmap +
                HTML_FOOTER)
    else:
        return hmap



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (C) Duncan Macleod (2014)
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""Fancy literalinclude to prepend lines with '>>>'
"""

from docutils.parsers.rst import directives
from sphinx.directives.code import LiteralInclude

from ... import version

__author__ = 'Duncan Macleod <duncan.macleod@ligo.org>'
__version__ = version.version


class GWpyLiteralInclude(LiteralInclude):
    """Fancy LiteralInclude to add doctest prefix '>>>' to all lines
    """
    def run(self):
        out = super(GWpyLiteralInclude, self).run()
        for block in out:
            if block.get('source', '').endswith('.py'):
                for i, child in enumerate(block.children):
                    content = []
                    for l in child.astext().splitlines():
                        if l.startswith('   '):
                            content.append('    %s' % l)
                        else:
                            content.append('>>> %s' % l)
                    block.children[i] = type(child)('\n'.join(content),
                                                    rawsource=child.rawsource)
                    block['language'] = 'python'
                block.rawsource = block.astext()
        return out

directives.register_directive('literalinclude', GWpyLiteralInclude)


def setup(*args, **kwargs):
    pass



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (C) Duncan Macleod (2014)
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""Convert GWpy example python files into rst files for sphinx documentation
"""

__author__ = 'Duncan Macleod <duncan.macleod@ligo.org'

import sys
import os
import argparse
import re

METADATA = {
    'author': 'sectionauthor',
    'currentmodule': 'currentmodule',
}


# -----------------------------------------------------------------------------
# parse command line

parser = argparse.ArgumentParser(description=__doc__)
parser.add_argument('infile', metavar='example.py',
                    help='python file to convert',)
parser.add_argument('outfile', metavar='example.rst', nargs='?',
                    help='rst file to write, default: print to screen')

args = parser.parse_args()

# -----------------------------------------------------------------------------
# parse python file

lines = open(args.infile, 'rb').read().splitlines()
output = []
header = []

indoc = False
incode = False


for i,line in enumerate(lines):
    # skip file header
    if len(output) == 0 and line.startswith('#'):
        continue

    # end on plot display
    if line.startswith(('if __name__ == ', '# Show')):
        break

    # hide lines
    if line.endswith('# hide'):
        continue

    # find block docs
    if line.startswith('"""'):
        indoc = not indoc
    line = line.strip('"')

    # skip empty lines not in a block quote
    if not line and not indoc:
        continue

    # find code
    if incode and line.startswith(('"', '#', '__')):
        incode = False
        output.append('')

    # find plots

    # comments
    if line.startswith('#'):
        output.append(line.strip('# '))
    # metadata
    elif line.startswith('__'):
        key, value = map(lambda x: x.strip(' _="\'').rstrip(' _="\''),
                         line.split('=', 1))
        if key in METADATA:
            header.append('.. %s:: %s\n' % (METADATA[key], value))
    # block quote
    elif indoc:
        output.append(line.strip('"').rstrip('"'))
    # code
    else:
        if not incode:
            output.append('')
        output.append('    >>> %s' % line)
        incode = True

    # end block quote
    if line.endswith('"""') and indoc:
        output.append('')
        indoc = False

    if len(output) == 1:
        output.append('#'*len(output[0]))

output.append('\n.. plot:: %s\n' % args.infile)
output = header + output

if args.outfile:
    with open(args.outfile, 'w') as f:
        f.write('\n'.join(output))
else:
    print('\n'.join(output))




# -*- coding: utf-8 -*-
# Copyright (C) Duncan Macleod (2013)
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""Extension of the basic matplotlib Figure for GWpy
"""

import numpy

from matplotlib import (backends, figure, pyplot, colors as mcolors,
                        _pylab_helpers)
from matplotlib.axes import SubplotBase
from matplotlib.cbook import iterable
from matplotlib.ticker import LogLocator

try:
    from mpl_toolkits.axes_grid1 import make_axes_locatable
except ImportError:
    from mpl_toolkits.axes_grid import make_axes_locatable

from . import (axes, utils)
from .axes import Axes
from .log import CombinedLogFormatterMathtext
from .decorators import (auto_refresh, axes_method)

__all__ = ['Plot']

try:
    __IPYTHON__
except NameError:
    IPYTHON = False
else:
    IPYTHON = True


def interactive_backend():
    return pyplot.get_backend() in backends.interactive_bk


class Plot(figure.Figure):
    """An extension of the core matplotlib :class:`~matplotlib.figure.Figure`.

    The `Plot` provides a number of methods to simplify generating
    figures from GWpy data objects, and modifying them on-the-fly in
    interactive mode.
    """
    _DefaultAxesClass = Axes

    def __init__(self, *args, **kwargs):
        # pull non-standard keyword arguments
        auto_refresh = kwargs.pop('auto_refresh', False)

        # generated figure, with associated interactivity from pyplot
        super(Plot, self).__init__(*args, **kwargs)
        backend_mod, _, draw_if_interactive, _ = backends.pylab_setup()
        manager = backend_mod.new_figure_manager_given_figure(1, self)
        cid = manager.canvas.mpl_connect(
            'button_press_event',
            lambda ev: _pylab_helpers.Gcf.set_active(manager))
        manager._cidgcf = cid
        _pylab_helpers.Gcf.set_active(manager)
        draw_if_interactive()

        # finalise
        self.set_auto_refresh(auto_refresh)
        self.colorbars = []
        self._coloraxes = []

    # -----------------------------------------------
    # core plot operations

    def get_auto_refresh(self):
        """Return this `Plot`s auto-refresh setting
        """
        return self._auto_refresh

    def set_auto_refresh(self, b):
        """Set this `Plot`s auto-refresh setting

        With auto_refresh set to `True`, all modifications of the underlying
        `Axes` will trigger the plot to be re-drawn

        Parameters
        ----------
        b : `True` or `False`
        """
        self._auto_refresh = bool(b)

    def refresh(self):
        """Refresh the current figure
        """
        for cbar in self.colorbars:
            cbar.draw_all()
        self.canvas.draw()

    def show(self, block=None, warn=True):
        """Display the current figure (if possible)

        Parameters
        ----------
        block : `bool`, default: `None`
            open the figure and block until the figure is closed, otherwise
            open the figure as a detached window. If `block=None`, GWpy
            will block if using an interactive backend and not in an
            ipython session.
        warn : `bool`, default: `True`
            if `block=False` is given, print a warning if matplotlib is
            not running in an interactive backend and cannot display the
            figure.

        Notes
        -----
        If blocking is employed, this method calls the
        :meth:`pyplot.show <matplotlib.pyplot.show>` function, otherwise
        the :meth:`~matplotlib.figure.Figure.show` method of this
        `~matplotlib.figure.Figure` is used.
        """
        # if told to block, or using an interactive backend,
        # but not using ipython
        if block or (block is None and interactive_backend() and not IPYTHON):
            return pyplot.show(block=True)
        # otherwise, don't block and just show normally
        else:
            return super(Plot, self).show(warn=warn)

    def save(self, *args, **kwargs):
        """Save the figure to disk.

        All `args` and `kwargs` are passed directly to the savefig
        method of the underlying `matplotlib.figure.Figure`
        self.fig.savefig(*args, **kwargs)
        """
        self.savefig(*args, **kwargs)

    def close(self):
        """Close the plot and release its memory.
        """
        for ax in self.axes:
            ax.cla()
        pyplot.close(self)

    # -----------------------------------------------
    # colour-bar

    @auto_refresh
    def add_colorbar(self, mappable=None, ax=None, location='right',
                     width=0.2, pad=0.1, log=None, label="", clim=None,
                     cmap=None, clip=None, visible=True, axes_class=axes.Axes,
                     **kwargs):
        """Add a colorbar to the current `Axes`

        Parameters
        ----------
        mappable : matplotlib data collection
            collection against which to map the colouring
        ax : :class:`~matplotlib.axes.Axes`
            axes from which to steal space for the colour-bar
        location : `str`, optional, default: 'right'
            position of the colorbar
        width : `float`, optional default: 0.2
            width of the colorbar as a fraction of the axes
        pad : `float`, optional, default: 0.1
            gap between the axes and the colorbar as a fraction of the axes
        log : `bool`, optional, default: `False`
            display the colorbar with a logarithmic scale
        label : `str`, optional, default: '' (no label)
            label for the colorbar
        clim : pair of floats, optional
            (lower, upper) limits for the colorbar scale, values outside
            of these limits will be clipped to the edges
        visible : `bool`, optional, default: `True`
            make the colobar visible on the figure, this is useful to
            make two plots, each with and without a colorbar, but
            guarantee that the axes will be the same size
        **kwargs
            other keyword arguments to be passed to the
            :meth:`~matplotlib.figure.Figure.colorbar` generator

        Returns
        -------
        Colorbar
            the :class:`~matplotlib.colorbar.Colorbar` added to this plot
        """
        # find default layer
        if mappable is None and ax is not None and len(ax.collections):
            mappable = ax.collections[-1]
        elif mappable is None and ax is not None and len(ax.images):
            mappable = ax.images[-1]
        elif (visible is False and mappable is None and
              ax is not None and len(ax.lines)):
            mappable = ax.lines[-1]
        elif mappable is None and ax is None:
            for ax in self.axes[::-1]:
                if hasattr(ax, 'collections') and len(ax.collections):
                    mappable = ax.collections[-1]
                    break
                elif hasattr(ax, 'images') and len(ax.images):
                    mappable = ax.images[-1]
                    break
                elif visible is False and len(ax.lines):
                    mappable = ax.lines[-1]
                    break
        if visible and mappable is None:
            raise ValueError("Cannot determine mappable layer for this "
                             "colorbar")
        elif ax is None:
            raise ValueError("Cannot determine an anchor Axes for this "
                             "colorbar")

        # find default axes
        if not ax:
            ax = mappable.axes

        mappables = ax.collections + ax.images

        # get new colour axis
        divider = make_axes_locatable(ax)
        if location not in ['right', 'top']:
            raise ValueError("'left' and 'bottom' colorbars have not "
                             "been implemented")
        cax = divider.append_axes(location, width, pad=pad,
                                  add_to_figure=visible, axes_class=axes_class)
        self._coloraxes.append(cax)
        if visible:
            self.sca(ax)
        else:
            return

        # set limits
        if not clim:
            clim = mappable.get_clim()
        if log is None:
            log = isinstance(mappable.norm, mcolors.LogNorm)
        if log and clim[0] <= 0.0:
            cdata = mappable.get_array()
            try:
                clim = (cdata[cdata > 0.0].min(), clim[1])
            except ValueError:
                pass
        for m in mappables:
            m.set_clim(clim)

        # set map
        if cmap is not None:
            mappable.set_cmap(cmap)

        # set normalisation
        norm = mappable.norm
        if clip is None:
            clip = norm.clip
        for m in mappables:
            if log and not isinstance(norm, mcolors.LogNorm):
                m.set_norm(mcolors.LogNorm(*mappable.get_clim()))
            elif not log:
                m.set_norm(mcolors.Normalize(*mappable.get_clim()))
            m.norm.clip = clip

        # set log ticks
        if log:
            kwargs.setdefault('ticks', LogLocator(subs=numpy.arange(1, 11)))
            kwargs.setdefault('format', CombinedLogFormatterMathtext())

        # make colour bar
        colorbar = self.colorbar(mappable, cax=cax, ax=ax, **kwargs)

        # set label
        if label:
            colorbar.set_label(label)
        colorbar.draw_all()

        self.colorbars.append(colorbar)
        return colorbar

    # -------------------------------------------
    # GWpy data adding
    #
    # These methods try to guess which axes to add to, otherwise generate
    # a new one

    def add_subplot(self, *args, **kwargs):
        kwargs.setdefault('projection', self._DefaultAxesClass.name)
        return super(Plot, self).add_subplot(*args, **kwargs)
    add_subplot.__doc__ = figure.Figure.add_subplot.__doc__

    def get_axes(self, projection=None):
        """Find all `Axes`, optionally matching the given projection

        Parameters
        ----------
        projection : `str`
            name of axes types to return
        """
        if projection is None:
            return self.axes
        else:
            return [ax for ax in self.axes if ax.name == projection.lower()]

    def _find_axes(self, projection=None):
        """Find the most recently added axes for the given projection

        Raises
        ------
        IndexError
            if no axes for the projection are found
        """
        try:
            return self.get_axes(projection)[-1]
        except IndexError:
            if projection:
                raise IndexError("No '%s' Axes found in this Plot"
                                 % projection)
            else:
                raise IndexError("No Axes found in this Plot")

    def _increment_geometry(self):
        """Try to determine the geometry to use for a new Axes

        Raises
        ------
        ValueError
            if geometry is ambiguous
        """
        if not len(self.axes):
            return (1, 1, 1)
        # get bottom axes
        try:
            ca = [ax for ax in self.axes if isinstance(ax, SubplotBase)][-1]
        except IndexError:
            ca = self.gca()
        current = ca.get_geometry()
        shape = current[:2]
        pos = current[2]
        num = shape[0] * shape[1]
        # if space left in this set
        if pos < num:
            return (shape[0], shape[1], pos+1)
        # or add a new column
        elif shape[1] > 1 and shape[0] == 1:
            return (1, shape[1] + 1, pos+1)
        # otherwise add a new row
        else:
            return (shape[0] + 1, 1, pos+1)

    def _add_new_axes(self, projection, **kwargs):
        # get new geomtry
        geometry = self._increment_geometry()
        # make new axes
        newax = self.add_subplot(*geometry, projection=projection, **kwargs)
        # update geometry for previous axes
        nrows = geometry[0]
        ncols = geometry[1]
        i = 0
        for ax in self.axes[:-1]:
            if isinstance(ax, SubplotBase):
                i += 1
                ax.change_geometry(nrows, ncols, i)
        return newax

    @auto_refresh
    def _plot(self, x, y, *args, **kwargs):
        """Add a line to the current plot

        Parameters
        ----------
        x : array-like
            x positions of the line points (in axis coordinates)
        y : array-like
            y positions of the line points (in axis coordinates)
        projection : `str`, optional, default: `'timeseries'`
            name of the Axes projection on which to plot
        ax : :class:`~gwpy.plotter.Axes`
            the `Axes` on which to add these data, if this is not given,
            a guess will be made as to the best `Axes` to use. If no
            appropriate axes are found, new `Axes` will be created
        newax : `bool`, optional, default: `False`
            force data to plot on a fresh set of `Axes`
        **kwargs
            additional keyword arguments passed directly on to
            the axes :meth:`~matplotlib.axes.Axes.plot` method.

        Returns
        -------
        Line2D
            the :class:`~matplotlib.lines.Line2D` for this line layer
        """
        # get axes options
        projection = kwargs.pop('projection', None)
        ax = kwargs.pop('ax', None)
        newax = kwargs.pop('newax', False)
        sharex = kwargs.pop('sharex', None)
        sharey = kwargs.pop('sharey', None)

        # set kwargs
        kwargs.setdefault("linestyle", "-")
        kwargs.setdefault("linewidth", 1)
        kwargs.setdefault("markersize", 0)

        # find relevant axes
        if ax is None and not newax:
            try:
                ax = self._find_axes(projection)
            except IndexError:
                newax = True
        if newax:
            ax = self._add_new_axes(projection=projection,
                                    sharex=sharex, sharey=sharey)
        # plot on axes
        return ax.plot(numpy.asarray(x), numpy.asarray(y), **kwargs)[0]

    @auto_refresh
    def _scatter(self, x, y, projection=None, ax=None, newax=False,
                 **kwargs):
        """Internal `Plot` method to scatter onto the most
        favourable `Axes`

        Parameters
        ----------
        x : array-like
            x positions of the line points (in axis coordinates)
        y : array-like
            y positions of the line points (in axis coordinates)
        projection : `str`, optional, default: `None`
            name of the Axes projection on which to plot
        ax : :class:`~gwpy.plotter.Axes`
            the `Axes` on which to add these data, if this is not given,
            a guess will be made as to the best `Axes` to use. If no
            appropriate axes are found, new `Axes` will be created
        newax : `bool`, optional, default: `False`
            force data to plot on a fresh set of `Axes`
        **kwargs.
            other keyword arguments passed to the
            :meth:`matplotlib.axes.Axes.scatter` function

        Returns
        -------
        Collection
            the :class:`~matplotlib.collections.Collection` for this
            scatter layer
        """
        # set kwargs
        kwargs.setdefault("s", 20)
        sharex = kwargs.pop('sharex', None)
        sharey = kwargs.pop('sharey', None)

        # find relevant axes
        if ax is None and not newax:
            try:
                ax = self._find_axes(projection)
            except IndexError:
                newax = True
        if newax:
            ax = self._add_new_axes(projection=projection,
                                    sharex=sharex, sharey=sharey)
        # plot on axes
        return ax.scatter(numpy.asarray(x), numpy.asarray(y), **kwargs)

    @auto_refresh
    def _imshow(self, image, projection=None, ax=None, newax=False, **kwargs):
        """Internal `Plot` method to imshow onto the most
        favourable `Axes`

        Parameters
        ----------
        x : array-like
            x positions of the line points (in axis coordinates)
        y : array-like
            y positions of the line points (in axis coordinates)
        projection : `str`, optional, default: `None`
            name of the Axes projection on which to plot
        ax : :class:`~gwpy.plotter.Axes`
            the `Axes` on which to add these data, if this is not given,
            a guess will be made as to the best `Axes` to use. If no
            appropriate axes are found, new `Axes` will be created
        newax : `bool`, optional, default: `False`
            force data to plot on a fresh set of `Axes`
        **kwargs.
            other keyword arguments passed to the
            :meth:`matplotlib.axes.Axes.imshow` function

        Returns
        -------
        Collection
            the :class:`~matplotlib.image.AxesImage` for this image
        """
        sharex = kwargs.pop('sharex', None)
        sharey = kwargs.pop('sharey', None)
        # find relevant axes
        if ax is None and not newax:
            try:
                ax = self._find_axes(projection)
            except IndexError:
                newax = True
        if newax:
            ax = self._add_new_axes(projection=projection,
                                    sharex=sharex, sharey=sharey)
        # plot on axes
        return ax.imshow(image, **kwargs)

    @auto_refresh
    def add_line(self, x, y, *args, **kwargs):
        """Add a line to the current plot

        Parameters
        ----------
        x : array-like
            x positions of the line points (in axis coordinates)
        y : array-like
            y positions of the line points (in axis coordinates)
        projection : `str`, optional, default: `None`
            name of the Axes projection on which to plot
        ax : :class:`~gwpy.plotter.Axes`
            the `Axes` on which to add these data, if this is not given,
            a guess will be made as to the best `Axes` to use. If no
            appropriate axes are found, new `Axes` will be created
        newax : `bool`, optional, default: `False`
            force data to plot on a fresh set of `Axes`
        **kwargs
            additional keyword arguments passed directly on to
            the axes :meth:`~matplotlib.axes.Axes.plot` method.

        Returns
        -------
        Line2D
            the :class:`~matplotlib.lines.Line2D` for this line layer
        """
        return self._plot(x, y, *args, **kwargs)

    @auto_refresh
    def add_scatter(self, x, y, **kwargs):
        """Add a set or points to the current plot

        Parameters
        ----------
        x : array-like
            x-axis data points
        y : array-like
            y-axis data points
        projection : `str`, optional, default: `None`
            name of the Axes projection on which to plot
        ax : :class:`~gwpy.plotter.Axes`
            the `Axes` on which to add these data, if this is not given,
            a guess will be made as to the best `Axes` to use. If no
            appropriate axes are found, new `Axes` will be created
        newax : `bool`, optional, default: `False`
            force data to plot on a fresh set of `Axes`
        **kwargs.
            other keyword arguments passed to the
            :meth:`matplotlib.axes.Axes.scatter` function

        Returns
        -------
        Collection
            the :class:`~matplotlib.collections.Collection` for this
            scatter layer
        """
        return self._scatter(x, y, **kwargs)

    @auto_refresh
    def add_image(self, image, projection=None, ax=None, newax=False,
                  **kwargs):
        """Add a 2-D image to this plot

        Parameters
        ----------
        image : `numpy.ndarray`
            2-D array of data for the image
        **kwargs
            other keyword arguments are passed to the
            :meth:`matplotlib.axes.Axes.imshow` function

        Returns
        -------
        image : :class:`~matplotlib.image.AxesImage`
        """
        return self._imshow(image, projection=projection, ax=ax, newax=newax,
                            **kwargs)

    @auto_refresh
    def add_timeseries(self, timeseries, projection='timeseries',
                       ax=None, newax=False, sharex=None, sharey=None,
                       **kwargs):
        """Add a :class:`~gwpy.timeseries.TimeSeries` trace to this plot

        Parameters
        ----------
        timeseries : :class:`~gwpy.timeseries.TimeSeries`
            the TimeSeries to display
        projection : `str`, optional, default: `'timeseries'`
            name of the Axes projection on which to plot
        ax : :class:`~gwpy.plotter.Axes`
            the `Axes` on which to add these data, if this is not given,
            a guess will be made as to the best `Axes` to use. If no
            appropriate axes are found, new `Axes` will be created
        newax : `bool`, optional, default: `False`
            force data to plot on a fresh set of `Axes`
        **kwargs
            other keyword arguments for the `Plot.add_line` function

        Returns
        -------
        Line2D
            the :class:`~matplotlib.lines.Line2D` for this line layer
        """
        return self.add_array(timeseries, 'timeseries', ax=ax, newax=newax,
                              sharex=sharex, sharey=sharey, **kwargs)

    @auto_refresh
    def add_spectrum(self, spectrum, projection='spectrum', ax=None,
                     newax=False, sharex=None, sharey=None, **kwargs):
        """Add a :class:`~gwpy.spectrum.core.Spectrum` trace to this plot

        Parameters
        ----------
        spectrum : :class:`~gwpy.spectrum.core.spectrum`
            the `Spectrum` to display
        projection : `str`, optional, default: `'Spectrum'`
            name of the Axes projection on which to plot
        ax : :class:`~gwpy.plotter.Axes`
            the `Axes` on which to add these data, if this is not given,
            a guess will be made as to the best `Axes` to use. If no
            appropriate axes are found, new `Axes` will be created
        newax : `bool`, optional, default: `False`
            force data to plot on a fresh set of `Axes`
        **kwargs
            other keyword arguments for the `Plot.add_line` function

        Returns
        -------
        Line2D
            the :class:`~matplotlib.lines.Line2D` for this line layer
        """
        return self.add_array(spectrum, 'spectrum', ax=ax, newax=newax,
                              sharex=sharex, sharey=sharey, **kwargs)

    @auto_refresh
    def add_spectrogram(self, spectrogram, projection='timeseries',
                        ax=None, newax=False, sharex=None, sharey=None,
                        **kwargs):
        """Add a :class:`~gwpy.spectrogram.core.Spectrogram` trace to
        this plot

        Parameters
        ----------
        spectrogram : :class:`~gwpy.spectrogram.core.Spectrogram`
            the `Spectrogram` to display
        projection : `str`, optional, default: `timeseries`
            name of the Axes projection on which to plot
        ax : :class:`~gwpy.plotter.Axes`
            the `Axes` on which to add these data, if this is not given,
            a guess will be made as to the best `Axes` to use. If no
            appropriate axes are found, new `Axes` will be created
        newax : `bool`, optional, default: `False`
            force data to plot on a fresh set of `Axes`
        **kwargs
            other keyword arguments for the `Plot.add_line` function

        Returns
        -------
        Line2D
            the :class:`~matplotlib.lines.Line2D` for this line layer
        """
        return self.add_array(spectrogram, 'timeseries', ax=ax, newax=newax,
                              sharex=sharex, sharey=sharey, **kwargs)

    @auto_refresh
    def add_array(self, array, projection, ax=None, newax=False,
                  sharex=None, sharey=None, **kwargs):
        """Add a :class:`~gwpy.data.array.Array` to this plot

        Parameters
        ----------
        array : :class:`~gwpy.data.array.Array`
            the `Array` to display
        projection : `str`
        ax : :class:`~gwpy.plotter.Axes`
            the `Axes` on which to add these data, if this is not given,
            a guess will be made as to the best `Axes` to use. If no
            appropriate axes are found, new `Axes` will be created
        newax : `bool`, optional, default: `False`
            force data to plot on a fresh set of `Axes`
        **kwargs
            other keyword arguments for the `Plot.add_line` function

        Returns
        -------
        Artist : :class:`~matplotlib.artist.Artist`
            the layer return from the relevant plotting function
        """
        # find relevant axes
        if ax is None and not newax:
            try:
                ax = self._find_axes(projection)
            except IndexError:
                newax = True
        if newax:
            ax = self._add_new_axes(projection=projection,
                                    sharex=sharex, sharey=sharey)
        # plot on axes
        return ax.plot(array, **kwargs)

    def add_dataqualityflag(self, flag, projection=None, ax=None, newax=False,
                            sharex=None, sharey=None, **kwargs):
        """Add a :class:`~gwpy.segments.flag.DataQualityFlag` to this plot

        Parameters
        ----------
        flag : :class:`~gwpy.segments.flag.DataQualityFlag`
            the `DataQualityFlag` to display

        """
        # find relevant axes
        if ax is None and not newax:
            try:
                ax = self._find_axes(projection)
            except IndexError:
                newax = True
        if newax:
            ax = self._add_new_axes(projection=projection,
                                    sharex=sharex, sharey=sharey)
        # plot on axes
        return ax.plot(flag, **kwargs)

    # -------------------------------------------
    # Plot legend

    @auto_refresh
    def add_legend(self, *args, **kwargs):
        """Add a legend to this `Plot` on the most favourable `Axes`

        All non-keyword `args` and `kwargs` are passed directly to the
        :meth:`~matplotlib.axes.Axes.legend` generator

        Returns
        -------
        Legend
            the :class:`~matplotlib.legend.Legend` for this plot
        """
        ax = kwargs.pop('ax', None)
        if ax is None:
            ax = self._find_axes()
        return ax.legend(*args, **kwargs)

    # -------------------------------------------
    # Convenience methods for single-axes plots
    #
    # The majority of methods in this section are decorated to call the
    # equivalent method of the current Axes, and so contain no actual code

    @axes_method
    def get_xlim(self):
        pass
    get_xlim.__doc__ = axes.Axes.get_xlim.__doc__

    @auto_refresh
    @axes_method
    def set_xlim(self, *args, **kwargs):
        pass
    set_xlim.__doc__ = axes.Axes.set_xlim.__doc__

    xlim = property(fget=get_xlim, fset=set_xlim,
                    doc='x-axis limits for the current axes')

    @axes_method
    def get_ylim(self):
        pass
    get_ylim.__doc__ = axes.Axes.get_ylim.__doc__

    @auto_refresh
    @axes_method
    def set_ylim(self, *args, **kwargs):
        pass
    set_ylim.__doc__ = axes.Axes.set_ylim.__doc__

    ylim = property(fget=get_ylim, fset=set_ylim,
                    doc='y-axis limits for the current axes')

    @axes_method
    def get_xlabel(self):
        pass
    get_xlabel.__doc__ = axes.Axes.get_xlabel.__doc__

    @axes_method
    @auto_refresh
    def set_xlabel(self, *args, **kwargs):
        pass
    set_xlabel.__doc__ = axes.Axes.set_xlabel.__doc__

    xlabel = property(fget=get_xlabel, fset=set_xlabel,
                      doc='x-axis label for the current axes')

    @axes_method
    def get_ylabel(self):
        pass
    get_ylabel.__doc__ = axes.Axes.get_ylabel.__doc__

    @auto_refresh
    @axes_method
    def set_ylabel(self, *args, **kwargs):
        pass
    set_ylabel.__doc__ = axes.Axes.set_ylabel.__doc__

    ylabel = property(fget=get_ylabel, fset=set_ylabel,
                      doc='y-axis label for the current axes')

    @axes_method
    def get_title(self):
        pass
    get_title.__doc__ = axes.Axes.get_title.__doc__

    @auto_refresh
    @axes_method
    def set_title(self, *args, **kwargs):
        pass
    set_title.__doc__ = axes.Axes.set_title.__doc__

    title = property(fget=get_title, fset=set_title,
                     doc='title for the current axes')

    @axes_method
    def get_xscale(self):
        pass
    get_xscale.__doc__ = axes.Axes.get_xscale.__doc__

    @auto_refresh
    @axes_method
    def set_xscale(self, *args, **kwargs):
        pass
    set_xscale.__doc__ = axes.Axes.set_xscale.__doc__

    @property
    def logx(self):
        """View x-axis in logarithmic scale
        """
        return self.get_xscale() == 'log'

    @logx.setter
    @auto_refresh
    def logx(self, log):
        if not self.logx and bool(log):
            self.set_xscale('log')
        elif not self.logx and bool(log):
            self.set_xscale('linear')

    @axes_method
    def get_yscale(self):
        pass
    get_yscale.__doc__ = axes.Axes.get_yscale.__doc__

    @auto_refresh
    @axes_method
    def set_yscale(self, *args, **kwargs):
        pass
    set_yscale.__doc__ = axes.Axes.set_yscale.__doc__

    @property
    def logy(self):
        """View y-axis in logarithmic scale
        """
        return self.get_yscale() == 'log'

    @logy.setter
    @auto_refresh
    def logy(self, log):
        if not self.logy and bool(log):
            self.set_yscale('log')
        elif not self.logy and bool(log):
            self.set_yscale('linear')

    @axes_method
    def html_map(self, filename, data=None, **kwargs):
        pass
    html_map.__doc__ = axes.Axes.html_map.__doc__

    # -----------------------------------------------
    # utilies

    @staticmethod
    def _get_axes_data(inputs, sep=False):
        """Determine the number of axes from the input args to this `Plot`

        Parameters
        ----------
        inputs : `list` of array-like data sets
            a list of data arrays, or a list of lists of data sets
        sep : `bool`, optional
            plot each set of data on a separate `Axes`

        Returns
        -------
        axesdata : `list` of lists of array-like data
            a `list` with one element per required `Axes` containing the
            array-like data sets for those `Axes`

        Notes
        -----
        The logic for this method is as follows:

        - if a `list` of data arrays are given, and `sep=False`, use 1 `Axes`
        - if a `list` of data arrays are given, and `sep=True`, use N `Axes,
          one for each data array
        - if a nested `list` of data arrays are given, ignore `sep` and
          use one `Axes` for each element in the top list.

        For example:

            >>> Plot._get_naxes([data1, data2], sep=False)
            [[data1, data2]]
            >>> Plot._get_naxes([data1, data2], sep=True)
            [[data1], [data2]]
            >>> Plot._get_naxes([[data1, data2], data3])
            [[data1, data2], [data3]]
        """
        # if not given list, default to 1
        if not iterable(inputs):
            return [inputs]
        elif not len(inputs):
            return []
        # if given a nested list of data, multiple axes are required
        if any([isinstance(x, (list, tuple, dict)) for x in inputs]):
            sep = True
        # build list of lists
        out = []
        if not sep:
            out.append([])
        for x in inputs:
            # if not sep, each element of inputs is a data set
            if not sep:
                out[0].append(x)
            # otherwise if this element is a list already, that's fine
            elif isinstance(x, (list, tuple)):
                out.append(x)
            elif isinstance(x, dict):
                out.append(x.values())
            else:
                out.append([x])
        return out

    @staticmethod
    def _parse_kwargs(kwargs):
        """Separate input kwargs dict into axes, and artist params

        The assumption is that all remaining kwargs should be passed to
        the underlying `Plot` constructor

        Parameters
        ----------
        kwargs : `dict`
            `dict` of input keyword arguments for `Plot`

        Returns
        -------
        axargs, artistargs : `list` of `dict`
            separated kwarg `dict` for the the axes and the artists
        """
        separatedargs = []
        for arglist in [utils.AXES_PARAMS, utils.ARTIST_PARAMS]:
            separatedargs.append(dict())
            for key in arglist:
                if key in kwargs:
                    separatedargs[-1][key] = kwargs.pop(key)
        return separatedargs



#!/usr/bin/env python

# Copyright (C) Duncan Macleod (2013)
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""Comparing the same `Channel` at different times

I'm interested in comparing the amplitude spectrum of a channel between a
known 'good' time - where the spectrum is what we expect it to be - and a
known 'bad' time - where some excess noise appeared and the spectrum
changed appreciably.
"""

__author__ = "Duncan Macleod <duncan.macleod@ligo.org>"
__currentmodule__ = 'gwpy.timeseries'

# First, we import the `TimeSeries`
from gwpy.timeseries import TimeSeries

# And we set the times of our investigation:
goodtime = 1061800700
badtime = 1061524816
duration = 120

# Next we :meth:`~TimeSeries.get` the data:
gooddata = TimeSeries.get('L1:PSL-ISS_PDB_OUT_DQ', goodtime, goodtime+duration)
baddata = TimeSeries.get('L1:PSL-ISS_PDB_OUT_DQ', badtime, badtime+duration)

# and calculate an `amplitude spectral density (ASD) <TimeSeries.asd>` using a 4-second Fourier transform with a 2-second overlap:
goodasd = gooddata.asd(4, 2)
badasd = baddata.asd(4, 2)

# Lastly, we make a plot of the data by `plotting <Spectrum.plot>` one `Spectrum`, and then adding the second:
plot = badasd.plot(label='Noisy data')
ax = plot.gca()
ax.plot(goodasd, label='Clean data')
ax.set_xlabel('Frequency [Hz]')
ax.set_xlim(10, 8000)
ax.set_ylabel(r'Noise ASD [1/$\sqrt{\mathrm{Hz}}$]')
ax.set_ylim(1e-6, 5e-4)
ax.grid(True, 'both', 'both')
plot.show()



#!/usr/bin/env python

# Copyright (C) Duncan Macleod (2013)
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>

"""Calculating and plotting a `Spectrum`

I'm interested in the level of ground motion surrounding a particular time
during commissioning of the Advanced LIGO Livingston Observatory. I don't
have access to the frame files on disk, so I'll need to use NDS.
"""

__author__ = "Duncan Macleod <duncan.macleod@ligo.org>"
__currentmodule__ = 'gwpy.spectrum'

# In order to generate a `Spectrum` we need to import the
# `~gwpy.timeseries.TimeSeries` and :meth:`~gwpy.timeseries.TimeSeries.get`
# the data:
from gwpy.timeseries import TimeSeries
lho = TimeSeries.get(
    'H1:LDAS-STRAIN,rds', 'August 1 2010', 'August 1 2010 00:02')
llo = TimeSeries.get(
    'L1:LDAS-STRAIN,rds', 'August 1 2010', 'August 1 2010 00:02')

# We can then call the :meth:`~gwpy.timeseries.TimeSeries.asd` method to
# calculated the amplitude spectral density for each
# `~gwpy.timeseries.TimeSeries`:
lhoasd = lho.asd(2, 1)
lloasd = llo.asd(2, 1)

# We can then :meth:`~Spectrum.plot` the spectra
plot = lhoasd.plot(color='b', label='LHO')
ax = plot.gca()
ax.plot(lloasd, color='g', label='LLO')
ax.set_xlim(40, 4096)
ax.set_ylim(1e-23, 7.5e-21)
ax.spines['right'].set_visible(False)
ax.spines['top'].set_visible(False)
ax.get_xaxis().tick_bottom()
ax.get_yaxis().tick_right()
plot.show()



# -*- coding: utf-8 -*-
# Copyright (C) Duncan Macleod (2013)
#
# This file is part of GWpy.

# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""User-friendly extensions to `glue.datafind`
"""

import os.path

from glue.lal import CacheEntry

from .. import version
from ..time import to_gps
from ..utils import with_import

__version__ = version.version
__author__ = 'Duncan Macleod <duncan.macleod@ligo.org>'


@with_import('glue.datafind')
def connect(host=None, port=None):
    """Open a new datafind connection

    Parameters
    ----------
    host : `str`
        name of datafind server to query
    port : `int`
        port of datafind server on host

    Returns
    -------
    connection : `~glue.datafind.GWDataFindHTTPConnection`
        the new open connection
    """
    port = port and int(port)
    if port is not None and port != 80:
        cert, key = datafind.find_credential()
        return datafind.GWDataFindHTTPSConnection(
            host=host, port=port, cert_file=cert, key_file=key)
    else:
        return datafind.GWDataFindHTTPConnection(host=host, port=port)


def find_frametype(channel, gpstime=None, frametype_match=None,
                   host=None, port=None, return_all=False, exclude_tape=False):
    """Find the frametype(s) that hold data for a given channel
    """
    from ..detector import Channel
    channel = Channel(channel)
    name = channel.name
    if gpstime is not None:
        gpstime = to_gps(gpstime).seconds
    connection = connect(host, port)
    types = connection.find_types(channel.ifo[0], match=frametype_match)
    # get reference frame for all types
    frames = []
    for ft in types:
        try:
            if gpstime is None:
                frame = connection.find_latest(
                    channel.ifo[0], ft, urltype='file')[0]
            else:
                frame = connection.find_frame_urls(
                    channel.ifo[0], ft, gpstime, gpstime, urltype='file',
                    on_gaps='ignore')[0]
        except (IndexError, RuntimeError):
            continue
        else:
            if os.access(frame.path, os.R_OK) and (
                    not exclude_tape or not on_tape(frame)):
                frames.append((ft, frame.path))
    # sort frames by allocated block size and regular size
    # (to put frames on tape at the bottom of the list)
    frames.sort(key=lambda x: (on_tape(x[1]), num_channels(x[1])))
    # search each frametype for the given channel
    found = []
    for ft, path in frames:
        if get_channel_type(name, path):
            if not return_all:
                return ft
            else:
                found.append(ft)
    if len(found) == 0 and gpstime:
        raise ValueError("Cannot locate %r in any known frametype at GPS=%d"
                         % (name, gpstime))
    elif len(found) == 0:
        raise ValueError("Cannot locate %r in any known frametype" % name)
    else:
        return found


@with_import('lalframe')
def num_channels(framefile):
    """Find the total number of channels in this framefile
    """
    frfile = lalframe.FrameUFrFileOpen(framefile, "r")
    frtoc = lalframe.FrameUFrTOCRead(frfile)
    return sum(
        getattr(lalframe, 'FrameUFrTOCQuery%sN' % type_.title())(frtoc) for
        type_ in ['adc', 'proc', 'sim'])


@with_import('lalframe')
def get_channel_type(channel, framefile):
    """Find the channel type in a given frame file

    Parameters
    ----------
    channel : `str`, `~gwpy.detector.Channel`
        name of data channel to find
    framefile : `str`
        path of GWF file in which to search

    Returns
    -------
    ctype : `str`
        the type of the channel ('adc', 'sim', or 'proc') if the
        channel exists in the table-of-contents for the given frame,
        otherwise `False`
    """
    name = str(channel)
    # read frame and table of contents
    frfile = lalframe.FrameUFrFileOpen(framefile, "r")
    frtoc = lalframe.FrameUFrTOCRead(frfile)
    for type_ in ['sim', 'proc', 'adc']:
        query = getattr(lalframe, 'FrameUFrTOCQuery%sName' % type_.title())
        i = 0
        while True:
            try:
                c = query(frtoc, i)
            except RuntimeError:
                break
            else:
                if c == name:
                    return type_
            i += 1
    return False


def find_best_frametype(channel, start, end, urltype='file',
                        host=None, port=None, allow_tape=True):
    """Intelligently select the best frametype from which to read this channel
    """
    start = to_gps(start).seconds
    end = to_gps(end).seconds
    frametype = find_frametype(channel, gpstime=start, host=host, port=port,
                               exclude_tape=not allow_tape)
    connection = connect(host=host, port=port)
    try:
        cache = connection.find_frame_urls(channel[0], frametype,
                                           start, end, urltype=urltype,
                                           on_gaps='error')
        if not allow_tape and on_tape(*cache):
            raise RuntimeError()
    except RuntimeError:
        alltypes = find_frametype(channel, gpstime=start, host=host, port=port,
                                  return_all=True, exclude_tape=not allow_tape)
        cache = [(ft, connection.find_frame_urls(
            channel[0], ft, start, end, urltype=urltype,
            on_gaps='ignore')) for ft in alltypes]
        if not allow_tape:
            cache = [ftc for ftc in cache if not on_tape(*ftc[1])]
        cache.sort(key=lambda x:
            len(x[1]) and -abs(x[1].to_segmentlistdict().values()[0]) or 0)
        try:
            return cache[0][0]
        except IndexError:
            raise ValueError("Cannot find any valid frametypes for %r"
                             % channel)
    else:
        return frametype


def on_tape(*files):
    """Determine whether any of the given files are on tape

    Parameters
    ----------
    *files : `str`, `~glue.lal.CacheEntry`
        one or more paths to GWF files

    Returns
    -------
    True/False : `bool`
        `True` if any of the files are determined to be on tape,
        otherwise `False`
    """
    for f in files:
        if isinstance(f, CacheEntry):
            f = f.path
        if os.stat(f).st_blocks == 0:
            return True
    return False



#!/usr/bin/env python

# Copyright (C) Duncan Macleod (2013)
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""Plotting a `Spectrogram` of the Rayleigh statistic

I would like to study the gravitational wave strain spectrogram around the time of an interesting simulated signal during the last science run (S6).
"""

__author__ = "Duncan Macleod <duncan.macleod@ligo.org>"
__currentmodule__ = 'gwpy.spectrogram'

# First, we import the :class:`~gwpy.timeseries.TimeSeries` and :meth:`~gwpy.timeseries.TimeSeries.get` the data:
from gwpy.timeseries import TimeSeries
gwdata = TimeSeries.get(
    'H1:LDAS-STRAIN,rds', 'September 16 2010 06:40', 'September 16 2010 06:50')

# Next, we can calculate a Rayleigh statistic `Spectrogram` using the 
# :meth:`~gwpy.timeseries.TimeSeries.rayleigh_spectrogram` method of the #
# `~gwpy.timeseries.TimeSeries` and a 5-second stride with a 2-second FFT and 
# 1-second overlap (50%):
rayleigh = gwdata.rayleigh_spectrogram(5, fftlength=2, overlap=1)

# and can make a plot using the :meth:`~Spectrogram.plot` method
plot = rayleigh.plot(norm='log', vmin=0.25, vmax=4)
plot.set_yscale('log')
plot.set_ylim(40, 4000)
plot.add_colorbar(label=r'Rayleigh statistic')
plot.show()
